#!/usr/bin/env python3

import unittest
from unittest.mock import mock_open, patch
from rkd.api.inputoutput import Wizard
from rkd.api.inputoutput import BufferedSystemIO
from rkd.test import TestTask
from rkd.exception import InterruptExecution


class TestWizard(unittest.TestCase):
    def test_regexp_validation(self):
        wizard = Wizard(TestTask())
        wizard.sleep_time = 0
        wizard.io = BufferedSystemIO()

        with self.subTest('Input is matching regexp'):
            wizard.input = lambda secret: 'Buenaventura'
            wizard.ask('What\'s your name?', 'name', regexp='([A-Za-z]+)')
            self.assertIn('name', wizard.answers)

        with self.subTest('Input is not matching regexp'):
            wizard.input = lambda secret: '...'
            self.assertRaises(InterruptExecution, lambda: wizard.ask('What\'s your name?', 'name', regexp='([A-Za-z]+)'))

    def test_input_is_retried_when_validation_fails(self):
        wizard = Wizard(TestTask())
        wizard.sleep_time = 0
        wizard.io = BufferedSystemIO()
        self.retry_num = 0

        def mocked_input(secret: bool = False):
            self.retry_num = self.retry_num + 1

            return ['...', '...', 'Buenaventura'][self.retry_num]

        wizard.input = mocked_input
        wizard.ask('What\'s your name?', 'name', regexp='([A-Za-z]+)')

        self.assertEqual({'name': 'Buenaventura'}, wizard.answers)

    def test_must_select_between_regexp_and_choice_validation(self):
        """Verify that the 'regexp' and 'choices' cannot be specified at one time in ask()"""

        wizard = Wizard(TestTask())
        wizard.io = BufferedSystemIO()

        self.assertRaises(Exception, lambda:
            wizard.ask('In which year the Spanish social revolution has begun?',
                       attribute='year',
                       regexp='([0-9]+)',
                       choices=['1936', '1910'])
        )

    def test_is_question_pretty_formatted_for_choice_validation(self):
        wizard = Wizard(TestTask())
        wizard.io = BufferedSystemIO()
        wizard.input = lambda secret: '1936'

        wizard.ask('In which year the Spanish social revolution has begun?',
                   attribute='year',
                   choices=['1936', '1910'])

        self.assertIn('In which year the Spanish social revolution has begun? [1936, 1910]:', wizard.io.get_value())

    def test_is_question_pretty_formatted_for_regexp_validation(self):
        wizard = Wizard(TestTask())
        wizard.io = BufferedSystemIO()
        wizard.input = lambda secret: '1936'

        wizard.ask('In which year the Spanish social revolution has begun?',
                   attribute='year',
                   regexp='([0-9]{4})')

        self.assertIn('In which year the Spanish social revolution has begun? [([0-9]{4})]:', wizard.io.get_value())

    def test_is_question_pretty_formatted_for_default_value_and_choice(self):
        wizard = Wizard(TestTask())
        wizard.io = BufferedSystemIO()
        wizard.input = lambda secret: '1936'

        wizard.ask('In which year the Spanish social revolution has begun?',
                   attribute='year',
                   regexp='([0-9]{4})',
                   default='1936')

        self.assertIn('In which year the Spanish social revolution has begun? [([0-9]{4})] [default: 1936]:', wizard.io.get_value())

    def test_finish_is_writing_values(self):
        tmp_wizard_file = mock_open()
        rkd_shell_calls = []

        with patch('rkd.api.inputoutput.open', tmp_wizard_file, create=True):
            task = TestTask()
            task.rkd = lambda *args, **kwargs: rkd_shell_calls.append(args)

            wizard = Wizard(task)
            wizard.io = BufferedSystemIO()

            wizard.input = lambda secret: '1936'
            wizard.ask('In which year the Spanish social revolution has begun?',
                       attribute='year',
                       regexp='([0-9]{4})',
                       default='1936')

            wizard.input = lambda secret: 'ait'
            wizard.ask('Enter new value for COMPOSE_PROJECT_NAME', attribute='COMPOSE_PROJECT_NAME', to_env=True)
            wizard.finish()

            # assertions
            tmp_wizard_file.assert_called_once_with('.rkd/tmp-wizard.json', 'wb')
            self.assertEqual([':env:set', '--name="COMPOSE_PROJECT_NAME"', '--value="ait"'], rkd_shell_calls[0][0])

    def test_getpass_is_used_when_secret_switch_is_used(self):
        """Test a secret=True switch in wizard.input() which is used by ask()"""

        wizard = Wizard(TestTask())
        wizard.io = BufferedSystemIO()

        with patch('rkd.api.inputoutput.getpass') as getpass:
            getpass.return_value = 'organize!'

            self.assertEqual('organize!', wizard.input(secret=True))



from typing import Dict
from argparse import ArgumentParser
from .api.syntax import TaskDeclaration
from .api.contract import TaskInterface
from .api.contract import ExecutionContext
from .standardlib import CallableTask
from .api.inputoutput import NullSystemIO
from .api.inputoutput import IO
from .context import ApplicationContext
from .execution.executor import OneByOneTaskExecutor
from .api.temp import TempManager


class TestTask(CallableTask):
    _description = 'Test task for automated tests'
    _become: str = False

    def __init__(self):
        self._io = NullSystemIO()

    def get_name(self) -> str:
        return ':test'

    def get_group_name(self) -> str:
        return ':rkd'

    def execute(self, context: ExecutionContext) -> bool:
        print('Hello world from :test task')
        return True

    def configure_argparse(self, parser: ArgumentParser):
        pass

    def get_declared_envs(self) -> Dict[str, str]:
        return {
            'ORG_NAME': 'International Workers Association'
        }


class TestTaskWithRKDCallInside(TestTask):
    def execute(self, context: ExecutionContext) -> bool:
        self.rkd([':sh', '-c', '"echo \'9 Aug 2014 Michael Brown, an unarmed Black teenager, was killed by a white police officer in Ferguson, Missouri, sparking mass protests across the US.\'"'])
        return True


def get_test_declaration(task: TaskInterface = None) -> TaskDeclaration:
    if not task:
        task = TestTask()

    return TaskDeclaration(task, {}, [])


def mock_task(task: TaskInterface, io: IO = None) -> TaskInterface:
    if io is None:
        io = NullSystemIO()

    ctx = ApplicationContext([], [], '')
    ctx.io = io

    task.internal_inject_dependencies(
        io=io,
        ctx=ctx,
        executor=OneByOneTaskExecutor(ctx),
        temp_manager=TempManager()
    )

    return task


def mock_execution_context(task: TaskInterface, args: Dict[str, str] = None, env: Dict[str, str] = None,
                           defined_args: Dict[str, dict] = None) -> ExecutionContext:
    if args is None: args = {}
    if env is None: env = {}
    if defined_args is None: defined_args = {}

    if args and not defined_args:
        for name, passed_value in args.items():
            defined_args[name] = {'default': ''}

    return ExecutionContext(
        TaskDeclaration(task),
        parent=None,
        args=args,
        env=env,
        defined_args=defined_args
    )


def ret_true() -> bool:
    return True


def ret_root() -> str:
    return 'root'


def ret_invalid_user() -> str:
    return 'invalid-user-there'




#!/usr/bin/env python3

import unittest
from rkd.aliasgroups import parse_alias_groups_from_env, AliasGroup


class AliasGroupsTest(unittest.TestCase):
    def test_parsed_count_matches(self):
        self.assertEqual(2, len(parse_alias_groups_from_env(':harbor->:hb,:harbor->')))
        self.assertEqual(0, len(parse_alias_groups_from_env('')))

    def test_resolves_from_empty_alias_destination_name(self):
        """Resolves :harbor:start into :harbor"""

        parsed = parse_alias_groups_from_env('->:harbor')
        self.assertEqual(':harbor:start', parsed[0].append_alias_to_task(':start'))

    def test_resolves_single_level_group(self):
        parsed = parse_alias_groups_from_env(':iwa->:international-workers-association')
        self.assertEqual(':international-workers-association:strike', parsed[0].append_alias_to_task(':iwa:strike'))

    def test_resolves_multiple_level_group(self):
        parsed = parse_alias_groups_from_env(':wl->:workers:liberation')
        self.assertEqual(':workers:liberation:strike', parsed[0].append_alias_to_task(':wl:strike'))

    def test_not_resolves_anything(self):
        parsed = parse_alias_groups_from_env(':iwa->:international-workers-association')
        self.assertEqual(None, parsed[0].append_alias_to_task(':capitalism:sucks'))

    def test_removes_previous_group_name(self):
        parsed = parse_alias_groups_from_env('->:harbor')
        self.assertEqual(':start', parsed[0].get_aliased_task_name(':harbor:start'))



#!/usr/bin/env python3

import os
import sys
import unittest
import tempfile
import subprocess
from copy import deepcopy
from tempfile import NamedTemporaryFile
from typing import Tuple
from contextlib import contextmanager
from io import StringIO
from rkd.api.inputoutput import IO
from rkd import RiotKitDoApplication

SCRIPT_DIR_PATH = os.path.dirname(os.path.realpath(__file__))


class TestFunctional(unittest.TestCase):
    """
    Functional tests case of the whole application.
    Runs application like from the shell, captures output and performs assertions on the results.
    """

    _stdout = None
    _stderr = None

    def setUp(self) -> None:
        os.environ['RKD_DEPTH'] = '0'
        self._stdout = sys.stdout
        self._stderr = sys.stderr

    def _restore_standard_out(self):
        sys.stdout = self._stdout
        sys.stderr = self._stderr

    @contextmanager
    def environment(self, environ: dict):
        backup = deepcopy(os.environ)

        try:
            os.environ.update(environ)
            yield
        finally:
            os.environ = backup

    def _run_and_capture_output(self, argv: list) -> Tuple[str, int]:
        io = IO()
        out = StringIO()
        exit_code = 0

        try:
            with io.capture_descriptors(stream=out, enable_standard_out=False):
                app = RiotKitDoApplication()
                app.main(['test_functional.py'] + argv)

        except SystemExit as e:
            self._restore_standard_out()
            exit_code = e.code

        return out.getvalue(), exit_code

    def test_tasks_listing(self):
        """ :tasks """

        full_output, exit_code = self._run_and_capture_output([':tasks'])

        self.assertIn(' >> Executing :tasks', full_output)
        self.assertIn('[global]', full_output)
        self.assertIn(':version', full_output)
        self.assertIn('succeed.', full_output)
        self.assertEqual(0, exit_code)

    def test_global_help_switch(self):
        """ --help """

        full_output, exit_code = self._run_and_capture_output(['--help'])

        self.assertIn('usage: :init', full_output)
        self.assertIn('--log-to-file', full_output)
        self.assertIn('--log-level', full_output)
        self.assertIn('--keep-going', full_output)
        self.assertIn('--silent', full_output)
        self.assertIn('--become', full_output)
        self.assertEqual(0, exit_code)

    def test_silent_switch_makes_tasks_task_to_not_show_headers(self):
        full_output, exit_code = self._run_and_capture_output([':tasks', '--silent'])

        # this is a global header
        self.assertIn(' >> Executing :tasks', full_output)

        # this is a header provided by :tasks
        self.assertNotIn('[global]', full_output)

        # the content is there
        self.assertIn(':exec', full_output)

    def test_global_silent_switch_is_making_silent_all_fancy_output(self):
        full_output, exit_code = self._run_and_capture_output(['--silent', ':tasks'])

        # content is there
        self.assertIn(':exec', full_output)

        # global formatting and per task - :tasks formatting is not there
        self.assertNotIn('>> Executing :tasks', full_output)   # global (SystemIO)
        self.assertNotIn('[global]', full_output)              # per-task (IO)

    def test_logging_tasks_into_separate_files(self):
        """Checks if RKD is able to log output to file per task
        """

        first = NamedTemporaryFile(delete=False)
        second = NamedTemporaryFile(delete=False)

        try:
            self._run_and_capture_output([
                ':version',
                '--log-to-file=' + first.name,

                ':tasks',
                '--log-to-file=' + second.name
            ])
        finally:
            # assertions
            with open(first.name) as first_handle:
                content = first_handle.read()

                self.assertIn('RKD version', content)  # RKD version globally as a tool
                self.assertIn(':sh', content)  # one of tasks
                self.assertIn('rkd.standardlib.VersionTask', content)

            with open(second.name) as second_handle:
                content = second_handle.read()

                self.assertIn(':exec', content)
                self.assertIn(':tasks', content)
                self.assertNotIn('>> Executing', content, msg='Global formatting should not be present')

                # assert that there is no output from previous task
                self.assertNotIn('RKD version', content)
                self.assertNotIn(':sh version', content)

                # clean up
            os.unlink(first.name)
            os.unlink(second.name)

    def test_env_variables_listed_in_help(self):
        full_output, exit_code = self._run_and_capture_output(['--help'])
        self.assertIn('- RKD_DEPTH (default: 0)', full_output)

    def test_env_variables_not_listed_in_sh_task(self):
        """ :sh does not define any environment variables """

        full_output, exit_code = self._run_and_capture_output([':sh', '--help'])
        self.assertNotIn('- RKD_DEPTH (default: )', full_output)
        self.assertIn('-- No environment variables declared --', full_output)

    def test_tasks_whitelist_shows_only_selected_groups(self):
        """Test that when we set RKD_WHITELIST_GROUPS=:rkd, then we will see only tasks from [rkd] group"""

        with self.environment({'RKD_WHITELIST_GROUPS': ':rkd'}):
            full_output, exit_code = self._run_and_capture_output([':tasks'])

        self.assertIn(':rkd:create-structure', full_output)
        self.assertNotIn(':exec', full_output)

    def test_task_whitelist_shows_only_global_group(self):
        """Test that when we set RKD_WHITELIST_GROUPS=,, then we will see only tasks from [global] group"""

        with self.environment({'RKD_WHITELIST_GROUPS': ','}):
            full_output, exit_code = self._run_and_capture_output([':tasks'])

        self.assertIn(':tasks', full_output)
        self.assertNotIn(':rkd:create-structure', full_output)

    def test_task_alias_resolves_task(self):
        """Test that with RKD_ALIAS_GROUPS=":py->:class-war" the :class-war:build would be resolved to :py:build"""

        with self.environment({'RKD_ALIAS_GROUPS': ':class-war->:rkd'}):
            full_output, exit_code = self._run_and_capture_output([':class-war:create-structure', '--help'])

        self.assertIn('usage: :rkd:create-structure', full_output)

    def test_env_variables_loaded_from_various_sources(self):
        """:hello task should print variables loaded globally and per-task using "environment" and "env_files"
        """

        with self.environment({'RKD_PATH': SCRIPT_DIR_PATH + '/../docs/examples/env-in-yaml/.rkd'}):
            full_output, exit_code = self._run_and_capture_output([':hello'])

            self.assertIn('Inline defined in this task: 17 May 1972 10,000 schoolchildren in the UK walked out on' +
                          ' strike in protest against corporal punishment. Within two years, London state schools ' +
                          'banned corporal punishment. The rest of the country followed in 1987.', full_output)

            self.assertIn('Inline defined globally: 16 May 1966, seamen across the UK walked out on a nationwide ' +
                          'strike for the first time in half a century. Holding solid for seven weeks, they won a' +
                          ' reduction in working hours from 56 to 48 per week', full_output)

            self.assertIn('Included globally - global.env: Jolanta Brzeska was a social activist against evictions,' +
                          ' she was murdered - burned alive by reprivatization mafia', full_output)

            self.assertIn('Included in task - per-task.env: 24 April 2013, the 8-storey Rana Plaza building in ' +
                          'Bangladesh collapsed, killing over 1,000 garment workers, as bosses in the ' +
                          'country\'s largest industry put profits before people', full_output)

    def test_help_shows_full_task_description(self):
        """:hello --help should show full description, even if it is multiline
        """

        with self.environment({'RKD_PATH': SCRIPT_DIR_PATH + '/../docs/examples/env-in-yaml/.rkd'}):
            full_output, exit_code = self._run_and_capture_output([':hello', '--help'])

            self.assertIn('Italian-American anarchist who was framed & executed', full_output)
            self.assertIn('#2 line: This is his short autobiography:', full_output)
            self.assertIn('#3 line: https://libcom.org/library/story-proletarian-life', full_output)

    def test_depth_increased(self):
        """Test that RKD_DEPTH is increased within next calls
        """

        full_output, exit_code = self._run_and_capture_output(
            [':sh', '-c', '%RKD% :sh -c \'echo "DEPTH: [$RKD_DEPTH]"\'']
        )

        self.assertIn('DEPTH: [2]', full_output)

    def test_depth_above_2_disables_ui(self):
        """Test that RKD in RKD will not show UI (fancy messages like "Executing ...")
        """

        full_output, exit_code = self._run_and_capture_output([
            '--no-ui', ':sh', '-c', '%RKD% :tasks'
        ])

        self.assertIn(':tasks', full_output)
        self.assertNotIn('>> Executing', full_output)

    def test_env_file_is_loaded_from_cwd(self):
        """Assert that .env file is loaded from current working directory
        """

        cwd_backup = os.getcwd()

        try:
            # 1. Create new working directory
            with tempfile.TemporaryDirectory() as tempdir:
                os.chdir(tempdir)

                # 2. Create example .env file
                with open(tempdir + '/.env', 'w') as f:
                    f.write("DURRUTI_BIRTHDAY_DATE=14.07.1896\n")

                # 3. Run
                full_output = subprocess.check_output(
                    sys.executable + " -m rkd :sh -c 'echo \"Durruti was born at $DURRUTI_BIRTHDAY_DATE\"'",
                    env={'PYTHONPATH': SCRIPT_DIR_PATH + '/../src'},
                    shell=True
                )

                # 4. Assert
                self.assertIn('Durruti was born at 14.07.1896', full_output.decode('utf-8'))
        finally:
            os.chdir(cwd_backup)



"""
    Tasks Executor
    ==============

    Responsible for executing tasks code, serialization, process separation, privileges management, passing the
    already prepared context, processing the results, processing declarative code

    Responsibilities are split into separate sub-layers.

"""



#!/usr/bin/env python3

import os
import subprocess
import unittest
from tempfile import TemporaryDirectory
from rkd.standardlib import CreateStructureTask
from rkd.test import mock_task, mock_execution_context
from rkd.api.inputoutput import BufferedSystemIO


class CreateStructureTaskTest(unittest.TestCase):
    @staticmethod
    def _execute_mocked_task(params: dict, envs: dict = {}, task: CreateStructureTask = None) -> BufferedSystemIO:
        io = BufferedSystemIO()

        if not task:
            task: CreateStructureTask = CreateStructureTask()

        mock_task(task, io=io)
        task.execute(mock_execution_context(task, params, envs))

        return io

    def test_functional_creates_structure_in_temporary_directory(self):
        """Test a successful case"""

        with TemporaryDirectory() as tempdir:
            cwd = os.getcwd()

            try:
                os.chdir(tempdir)
                task = CreateStructureTask()
                task.get_rkd_version_selector = lambda: ''

                self._execute_mocked_task({
                    '--commit': False,
                    '--no-venv': False,
                    '--pipenv': False
                }, {}, task=task)

                self.assertTrue(os.path.isdir(tempdir + '/.rkd'),
                                msg='Expected that .rkd directory would be created')
                self.assertTrue(os.path.isfile(tempdir + '/requirements.txt'),
                                msg='Expected requirements.txt file to be present')
                self.assertTrue(os.path.isfile(tempdir + '/.venv/bin/activate'),
                                msg='Expected that virtual environment will contain a bin/activate file')
                self.assertFalse(os.path.isfile(tempdir + '/Pipfile'),
                                 msg='Expected that pipenv structure would not be created ' +
                                     'if --pipenv switch not used explicitly')

            finally:
                os.chdir(cwd)

    def test_functional_detects_git_is_dirty(self):
        """Verify that GIT workspace is unclean - there are pending working changes not commited"""

        with TemporaryDirectory() as tempdir:
            cwd = os.getcwd()

            try:
                os.chdir(tempdir)

                # ---------------------------
                # Prepare the workspace first
                # ---------------------------
                subprocess.call(['git', 'init'])
                subprocess.call('echo "test-1" > test-file.txt; git add test-file.txt', shell=True)
                subprocess.call('git commit -m "First commit"', shell=True)
                # make the working tree dirty by editing a file without a commit
                subprocess.call('echo "changed" > test-file.txt', shell=True)

                task = CreateStructureTask()
                task.get_rkd_version_selector = lambda: ''
                io = self._execute_mocked_task({
                    '--commit': True,
                    '--no-venv': False,
                    '--pipenv': False
                }, {}, task=task)

                self.assertIn('Current working directory is dirty', io.get_value())
            finally:
                os.chdir(cwd)

    def test_functional_no_virtualenv_environment_is_created_when_a_switch_was_used(self):
        """Verify that --no-venv switch would end with skipped virtual environment creation"""

        with TemporaryDirectory() as tempdir:
            cwd = os.getcwd()

            try:
                os.chdir(tempdir)

                # action
                self._execute_mocked_task({'--commit': False, '--no-venv': True, '--pipenv': False}, {})

                # assertions
                self.assertTrue(os.path.isfile(tempdir + '/requirements.txt'),
                                msg='Expected requirements.txt file to be present')
                self.assertFalse(os.path.isfile(tempdir + '/.venv/bin/activate'))

            finally:
                os.chdir(cwd)

    def test_functional_interface_methods_are_called(self):
        """Verify that all defined extensible interface methods are called"""

        with TemporaryDirectory() as tempdir:
            cwd = os.getcwd()

            try:
                os.chdir(tempdir)
                subprocess.call(['git', 'init'])

                # verification
                call_history = []

                # mock
                task = CreateStructureTask()
                # do not set fixed version, as on local environment it could be some dev version not released yet
                task.get_rkd_version_selector = lambda: ''

                task.on_startup = lambda ctx: call_history.append('on_startup')
                task.on_files_copy = lambda ctx: call_history.append('on_files_copy')
                task.on_requirements_txt_write = lambda ctx: call_history.append('on_requirements_txt_write')
                task.on_creating_venv = lambda ctx: call_history.append('on_creating_venv')
                task.on_git_add = lambda ctx: call_history.append('on_git_add')
                task.get_patterns_to_add_to_gitignore = lambda ctx: [
                    'lets-ignore-all-the-opposites-of-fate-and-rise-up',
                    '.venv-setup.log'
                ]

                # action
                self._execute_mocked_task({'--commit': True, '--no-venv': False, '--pipenv': False}, {}, task=task)

                # assert that actions will be called in order

                with self.subTest('First time should copy files, create virtual env'):
                    self.assertEqual(
                        ['on_startup', 'on_files_copy', 'on_requirements_txt_write', 'on_creating_venv', 'on_git_add'],
                        call_history
                    )

                #
                # Reset to perform a test checking, that any next same task execution results in skipped files copying
                #
                call_history = []
                subprocess.check_call('echo "new" > new-file.txt; git add new-file.txt', shell=True)

                self._execute_mocked_task({'--commit': True, '--no-venv': False, '--pipenv': False}, {}, task=task)

                with self.subTest('Any next time should not copy files over and over again'):
                    self.assertEqual(
                        ['on_startup', 'on_requirements_txt_write', 'on_creating_venv', 'on_git_add'],
                        call_history
                    )

            finally:
                os.chdir(cwd)

    def test_pipenv_is_supported(self):
        """Check that pipenv structure is created"""

        with TemporaryDirectory() as tempdir:
            cwd = os.getcwd()

            try:
                os.chdir(tempdir)
                task = CreateStructureTask()
                task.get_rkd_version_selector = lambda: ''

                io = self._execute_mocked_task({
                    '--commit': False,
                    '--no-venv': False,
                    '--pipenv': True
                }, {}, task=task)

                self.assertTrue(os.path.isdir(tempdir + '/.rkd'),
                                msg='Expected that .rkd directory would be created')
                self.assertTrue(os.path.isfile(tempdir + '/requirements.txt'),
                                msg='Expected requirements.txt file to be present')
                self.assertFalse(os.path.isfile(tempdir + '/.venv/bin/activate'),
                                msg='Expected that a normal virtual env will not be created')
                self.assertTrue(os.path.isfile(tempdir + '/Pipfile'), msg='Expected a Pipfile created by pipenv')

                self.assertTrue(os.path.isfile(tempdir + '/Pipfile.lock'), msg='Expected a Pipfile.lock created by pipenv')

                self.assertIn('Structure created, use "pipenv shell" to enter project environment', io.get_value(),
                              msg='Expected a welcome message / instruction')

            finally:
                os.chdir(cwd)



#!/usr/bin/env python3

import unittest
import os
import yaml
from io import StringIO
from rkd.yaml_context import YamlSyntaxInterpreter
from rkd.yaml_parser import YamlFileLoader
from rkd.api.inputoutput import IO, NullSystemIO, BufferedSystemIO
from rkd.exception import DeclarationException, YamlParsingException
from rkd.contract import ExecutionContext

SCRIPT_DIR_PATH = os.path.dirname(os.path.realpath(__file__))


class TestYamlContext(unittest.TestCase):
    def test_parse_imports_successful_case_single_task(self):
        factory = YamlSyntaxInterpreter(NullSystemIO(), YamlFileLoader([]))
        imported = factory.parse_imports(['rkd.standardlib.jinja.RenderDirectoryTask'])

        self.assertEqual(':j2:directory-to-directory', imported[0].to_full_name())

    def test_parse_imports_successful_case_module(self):
        factory = YamlSyntaxInterpreter(NullSystemIO(), YamlFileLoader([]))
        imported = factory.parse_imports(['rkd.standardlib.jinja'])

        names_of_imported_tasks = []

        for task in imported:
            names_of_imported_tasks.append(task.to_full_name())

        self.assertIn(':j2:render', names_of_imported_tasks)
        self.assertIn(':j2:directory-to-directory', names_of_imported_tasks)

    def test_parse_imports_wrong_class_type_but_existing(self):
        def test():
            factory = YamlSyntaxInterpreter(NullSystemIO(), YamlFileLoader([]))
            factory.parse_imports(['rkd.exception.ContextException'])

        self.assertRaises(DeclarationException, test)

    def test_parse_imports_cannot_import_non_existing_class(self):
        def test():
            factory = YamlSyntaxInterpreter(NullSystemIO(), YamlFileLoader([]))
            factory.parse_imports(['rkd.standardlib.python.WRONG_NAME'])

        self.assertRaises(YamlParsingException, test)

    def test_parse_imports_importing_whole_module_without_submodules(self):
        factory = YamlSyntaxInterpreter(NullSystemIO(), YamlFileLoader([]))
        imported = factory.parse_imports(['rkd_python'])

        names_of_imported_tasks = []

        for task in imported:
            names_of_imported_tasks.append(task.to_full_name())

        self.assertIn(':py:build', names_of_imported_tasks)
        self.assertIn(':py:publish', names_of_imported_tasks)

    def test_parse_tasks_successful_case(self):
        """Successful case with description, arguments and bash steps
        """

        input_tasks = {
            ':resistentia': {
                'description': 'Against moving the costs of the crisis to the workers!',
                'arguments': {
                    '--picket': {
                        'help': 'Picket form',
                        'required': False,
                        'action': 'store_true'
                    }
                },
                'steps': [
                    'echo "Resistentia!"'
                ]
            }
        }

        io = IO()
        out = StringIO()
        factory = YamlSyntaxInterpreter(io, YamlFileLoader([]))
        parsed_tasks = factory.parse_tasks(input_tasks, '', './makefile.yaml', {})

        self.assertEqual(':resistentia', parsed_tasks[0].to_full_name(),
                         msg='Expected that the task name will be present')

        declaration = parsed_tasks[0]
        declaration.get_task_to_execute()._io = NullSystemIO()

        with io.capture_descriptors(stream=out, enable_standard_out=False):
            declaration.get_task_to_execute().execute(ExecutionContext(declaration))

        self.assertIn('Resistentia!', out.getvalue(), msg='Expected that echo contents will be visible')

    def test_parse_tasks_signals_error_instead_of_throwing_exception(self):
        """
        Test that error thrown by executed Python code will
        """

        input_tasks = {
            ':song': {
                'description': 'Bella Ciao is an Italian protest folk song that originated in the hardships of the ' +
                               'mondina women, the paddy field workers in the late 19th century who sang it to ' +
                               'protest against harsh working conditions in the paddy fields of North Italy',
                'steps': [
                    '''#!python
print(syntax-error-here)
                    '''
                ]
            }
        }

        io = BufferedSystemIO()
        factory = YamlSyntaxInterpreter(io, YamlFileLoader([]))
        parsed_tasks = factory.parse_tasks(input_tasks, '', 'makefile.yaml', {})

        declaration = parsed_tasks[0]
        declaration.get_task_to_execute()._io = IO()
        task = declaration.get_task_to_execute()
        task._io = io

        # execute prepared task
        result = task.execute(ExecutionContext(declaration))

        self.assertEqual(False, result, msg='Expected that syntax error would result in a failure')
        self.assertIn("NameError: name 'syntax' is not defined", io.get_value(), msg='Error message should be attached')
        self.assertIn('File ":song@step 1", line 1', io.get_value(), msg='Stacktrace should be attached')

    def test_parse_env_parses_environment_variables_added_manually(self):
        """Test "environment" block
        """

        io = IO()
        factory = YamlSyntaxInterpreter(io, YamlFileLoader([]))
        envs = factory.parse_env({
            'environment': {
                'EVENT_NAME': 'In memory of Maxwell Itoya, an Nigerian immigrant killed by police at flea market.' +
                              ' He was innocent, and left wife with 3 kids.'
            }
        }, 'make/file/path/makefile.yaml')

        self.assertIn('EVENT_NAME', envs)
        self.assertIn('Maxwell Itoya', envs['EVENT_NAME'])

    def test_parse_env_parses_environment_variables_from_file(self):
        """Check if envs are loaded from file correctly
        """

        io = IO()
        factory = YamlSyntaxInterpreter(io, YamlFileLoader([]))
        envs = factory.parse_env({
            'env_files': [
                SCRIPT_DIR_PATH + '/../docs/examples/env-in-yaml/.rkd/env/global.env'
            ]
        }, SCRIPT_DIR_PATH + '/../docs/examples/env-in-yaml/.rkd/makefile.yml')

        self.assertIn('TEXT_FROM_GLOBAL_ENV', envs)
        self.assertIn('Jolanta Brzeska was a social activist against evictions, ' +
                      'she was murdered - burned alive by reprivatization mafia', envs['TEXT_FROM_GLOBAL_ENV'])

    def test_parse_env_preserves_variables_order(self):
        """Make sure that the environment variables are loaded in order they were defined
        """

        yaml_content = '''
environment:
    FIRST:  "Jolanta Brzeska"
    SECOND: "Maxwell Itoya"
    THIRD:  "August Spies"
    FOURTH: "Samuel Fielden"
        '''

        expected_order = [
            "Jolanta Brzeska",
            "Maxwell Itoya",
            "August Spies",
            "Samuel Fielden"
        ]

        for i in range(1, 10000):
            parsed = yaml.load(yaml_content, yaml.FullLoader)

            io = IO()
            factory = YamlSyntaxInterpreter(io, YamlFileLoader([]))
            envs = factory.parse_env(parsed, 'makefile.yaml')

            names_in_order = []

            for env_name, value in envs.items():
                names_in_order.append(env_name)

            self.assertEqual(expected_order, list(envs.values()))



#!/usr/bin/env python3

import unittest
import unittest.mock
import os
import psutil
import subprocess
from tempfile import NamedTemporaryFile
from collections import OrderedDict
from io import StringIO
from rkd.standardlib import InitTask
from rkd.api.inputoutput import IO

CURRENT_SCRIPT_PATH = os.path.dirname(os.path.realpath(__file__))


class TestTaskUtil(unittest.TestCase):
    def test_sh_accepts_script_syntax(self):
        task = InitTask()
        task._io = IO()
        self.assertIn('__init__.py', task.sh("ls -la\npwd", capture=True))

    def test_exec_spawns_process(self):
        task = InitTask()
        task._io = IO()
        self.assertIn('__init__.py', task.exec('ls', capture=True))

    def test_sh_executes_in_background(self):
        task = InitTask()
        task._io = IO()
        task.exec('ls', background=True)

    def test_exec_background_capture_validation_raises_error(self):
        def test():
            task = InitTask()
            task._io = IO()
            task.exec('ls', background=True, capture=True)

        self.assertRaises(Exception, test)

    def test_sh_captures_output_in_correct_order_with_various_timing(self):
        """Test if output is containing stdout and stderr lines mixed in proper order (as it is defined in shell script)

        Notice: Test is interacting with shell, to reduce possibility of weird behavior it is retried multiple times
        """
        for i in range(1, 100):
            self.maxDiff = None  # unittest setting
            task = InitTask()
            task._io = IO()

            io = IO()
            out = StringIO()

            with io.capture_descriptors(stream=out, enable_standard_out=False):
                task.sh(''' set +e;
                    sleep 0.05;
                    echo "FIRST";
                    sleep 0.05;
                    echo "SECOND" >&2;
                    echo "THIRD";
                    echo "FOURTH" >&2;
                    echo "FIFTH" >&2;
                    echo "SIXTH";
                    echo "SEVENTH" >&2;
                    echo "NINETH";
                    echo "TENTH";
                ''')

            self.assertEqual("FIRST\nSECOND\nTHIRD\nFOURTH\nFIFTH\nSIXTH\nSEVENTH\nNINETH\nTENTH\n", out.getvalue())

    def test_sh_producing_large_outputs(self):
        """Process a few megabytes of output and assert that:

        - It will consume not more than 10 megabytes (assuming also output capturing in tests by io.capture_descriptors())
        - The whole output would be printed correctly
        """

        self.maxDiff = None  # unittest setting
        task = InitTask()
        task._io = IO()

        io = IO()
        out = StringIO()
        text = "History isn't made by kings and politicians, it is made by us."

        memory_before = psutil.Process(os.getpid()).memory_info().rss / 1024 / 1024

        with io.capture_descriptors(stream=out, enable_standard_out=False):
            task.py('''
for i in range(0, 1024 * 128):
    print("''' + text + '''")
            ''')

        iterations = 1024 * 128
        text_with_newlines_length = len(text) + 1
        memory_after = psutil.Process(os.getpid()).memory_info().rss / 1024 / 1024

        self.assertEqual(iterations * text_with_newlines_length, len(out.getvalue()))
        self.assertLessEqual(memory_after - memory_before, 15, msg='Expected less than 15 megabytes of memory usage')

    def test_sh_captures_output_in_correct_order_with_fixed_timing(self):
        """Test if output contains stdout and stderr lines printed out in proper order,
        while there is a sleep between prints

        Notice: Test is interacting with shell, to reduce possibility of weird behavior it is retried multiple times
        """

        for i in range(1, 30):
            self.maxDiff = None  # unittest setting
            task = InitTask()
            task._io = IO()

            io = IO()
            out = StringIO()

            with io.capture_descriptors(stream=out, enable_standard_out=False):
                task.sh(''' set +e;
                    sleep 0.05;
                    echo "FIRST";
                    sleep 0.05;
                    echo "SECOND" >&2;
                    sleep 0.05;
                    echo "THIRD";
                ''')

            self.assertEqual("FIRST\nSECOND\nTHIRD\n", out.getvalue())

    def test_sh_rkd_in_rkd_shows_first_lines_on_error(self):
        """Bugfix: sh() was loosing first line(s) of output, when exception was raised

        Notice: Test is interacting with shell, to reduce possibility of weird behavior it is retried multiple times
        """

        for i in range(1, 5):
            for std_redirect in ['', '>&2']:
                task = InitTask()
                task._io = IO()
                io = IO()
                out = StringIO()

                with io.capture_descriptors(stream=out, enable_standard_out=False):
                    try:
                        task.sh(''' 
                            python3 -m rkd --silent :sh -c 'echo "Bartolomeo Vanzetti" ''' + std_redirect + '''; exit 127'
                        ''')
                    except subprocess.CalledProcessError:
                        pass

                self.assertIn('Bartolomeo Vanzetti', out.getvalue(),
                              msg='Expected that output will be shown for std_redirect=%s' % std_redirect)

    def test_sh_3rd_depth_rkd_calls(self):
        """Bugfix: sh() of 3-depth calls -> test -> rkd -> rkd returns first line of output
        """

        for std_redirect in ['', '>&2']:
            task = InitTask()
            task._io = IO()
            io = IO()
            out = StringIO()

            with io.capture_descriptors(stream=out, enable_standard_out=False):
                try:
                    task.sh(''' 
                        python3 -m rkd --silent :sh -c 'python -m rkd --silent :sh -c \'echo "Nicola Sacco" ''' + std_redirect + '''; exit 1\''
                    ''')
                except subprocess.CalledProcessError:
                    pass

            self.assertIn('Nicola Sacco', out.getvalue(),
                          msg='Expected that output will be shown for std_redirect=%s' % std_redirect)

    @unittest.skip("Github issue #1")
    def test_sh_provides_stdout_and_stderr_in_exception(self):
        task = InitTask()
        task._io = IO()

        try:
            task.sh('''
                echo "Freedom, Equality, Mutual Aid!"
                echo "The state and capitalism is failing" >&2

                exit 161
            ''')
        except subprocess.CalledProcessError as e:
            self.assertEqual("Freedom, Equality, Mutual Aid!\n", e.output)
            self.assertEqual("The state and capitalism is failing\n", e.stderr)

    def test_sh_has_valid_order_of_defining_environment_variables(self):
        """Checks if built-in sh() method is registering variables in proper order
        """

        task = InitTask()
        task._io = IO()

        envs = OrderedDict()
        envs['ENV_NAME'] = 'riotkit'
        envs['COMPOSE_CMD'] = 'docker-compose -p ${ENV_NAME}'

        out = task.sh('''
            echo "${COMPOSE_CMD} up -d"
        ''', env=envs, capture=True)

        self.assertEqual('docker-compose -p riotkit up -d', out.strip())

    def test_py_executes_python_scripts_without_specifying_script_path(self):
        """Simply - check basic successful case - executing a Python code"""

        task = InitTask()
        task._io = IO()
        out = task.py('''
import os
print(os)
        ''', capture=True)

        self.assertIn("<module 'os' from", out)

    def test_py_executes_a_custom_python_script(self):
        """Check that script from specified file in 'script_path' parameter will be executed
        And the code will be passed to that script as stdin.
        """

        task = InitTask()
        task._io = IO()

        with NamedTemporaryFile() as temp_file:
            temp_file.write(b'import sys; print("STDIN: " + str(sys.stdin.read()))')
            temp_file.flush()

            out = task.py('Hello!', capture=True, script_path=temp_file.name)

        self.assertEqual('STDIN: Hello!\n', out)

    def test_py_inherits_environment_variables(self):
        os.putenv('PY_INHERITS_ENVIRONMENT_VARIABLES', 'should')

        task = InitTask()
        task._io = IO()
        out = task.py(
            code='import os; print("ENV VALUE IS: " + str(os.environ["PY_INHERITS_ENVIRONMENT_VARIABLES"]))',
            capture=True
        )

        self.assertEqual('ENV VALUE IS: should\n', out)

    def test_py_uses_sudo_when_become_specified(self):
        """Expect that sudo with proper parameters is used"""

        task = InitTask()
        task._io = IO()

        with unittest.mock.patch('rkd.taskutil.check_output') as mocked_subprocess:
            mocked_subprocess: unittest.mock.MagicMock
            task.py(code='print("test")', capture=True, become='root')

        self.assertEqual('sudo -E -u root python', mocked_subprocess.call_args[0][0])




"""Application container - manages a list of available tasks to execute (all imported)
"""

import os
import sys
from datetime import datetime
from typing import Dict, List, Union, Tuple
from importlib.machinery import SourceFileLoader
from traceback import print_exc
from .api.syntax import TaskDeclaration
from .api.syntax import TaskAliasDeclaration
from .api.syntax import GroupDeclaration
from .api.contract import ContextInterface
from .argparsing import CommandlineParsingHelper
from .api.inputoutput import SystemIO
from .exception import TaskNotFoundException
from .exception import ContextFileNotFoundException
from .exception import PythonContextFileNotFoundException
from .exception import NotImportedClassException
from .exception import ContextException
from .yaml_context import YamlSyntaxInterpreter
from .yaml_parser import YamlFileLoader


CURRENT_SCRIPT_PATH = os.path.dirname(os.path.realpath(__file__))


class ApplicationContext(ContextInterface):
    """
    Application context - collects all tasks together

    Each Context() is collecting tasks from selected directory.
    merge() static method is merging two Context() objects
    eg. merge(Context(), Context()) selecting second as a priority

    ContextFactory() is controlling the order.
    """

    _imported_tasks: Dict[str, TaskDeclaration]
    _task_aliases: Dict[str, TaskAliasDeclaration]
    _compiled: Dict[str, Union[TaskDeclaration, GroupDeclaration]]
    _created_at: datetime
    _directory: str
    directories: []
    io: SystemIO

    def __init__(self, tasks: List[TaskDeclaration], aliases: List[TaskAliasDeclaration], directory: str):
        self._imported_tasks = {}
        self._task_aliases = {}
        self._created_at = datetime.now()
        self._directory = directory
        self.directories = [directory] if directory else []

        for task in tasks:
            self._add_component(task)

        for alias in aliases:
            self._add_task(alias)

    @classmethod
    def merge(cls, first, second):
        """ Add one context to other context. Produces immutable new context. """

        new_ctx = cls([], [], '')

        for context in [first, second]:
            context: ApplicationContext

            for name, component in context._imported_tasks.items():
                new_ctx._add_component(component)

            for name, task in context._task_aliases.items():
                new_ctx._add_task(task)

            new_ctx.directories += context.directories

        return new_ctx

    def compile(self) -> None:
        """ Resolve all objects in the context. Should be called only, when all contexts were merged """

        self._compiled = self._imported_tasks

        for name, details in self._task_aliases.items():
            self._compiled[name] = self._resolve_alias(name, details)

    def find_task_by_name(self, name: str) -> Union[TaskDeclaration, GroupDeclaration]:
        try:
            return self._compiled[name]
        except KeyError:
            raise TaskNotFoundException(('Task "%s" is not defined. Check if it is defined, or' +
                                         ' imported, or if the spelling is correct.') % name)

    def find_all_tasks(self) -> Dict[str, Union[TaskDeclaration, GroupDeclaration]]:
        return self._compiled

    def get_creation_date(self) -> datetime:
        return self._created_at

    def _add_component(self, component: TaskDeclaration) -> None:
        self._imported_tasks[component.to_full_name()] = component

    def _add_task(self, task: TaskAliasDeclaration) -> None:
        self._task_aliases[task.get_name()] = task

    def _resolve_alias(self, name: str, alias: TaskAliasDeclaration) -> GroupDeclaration:
        """
        Parse commandline args to fetch list of tasks to join into a group

        Produced result will be available to fetch via find_task_by_name()
        """

        args = CommandlineParsingHelper.create_grouped_arguments(alias.get_arguments())
        resolved_tasks = []

        for argument_group in args:
            resolved_task = self.find_task_by_name(argument_group.name())

            # preserve original task env, and append alias env in priority
            merged_env = resolved_task.get_env()
            merged_env.update(alias.get_env())

            new_task = resolved_task \
                .with_env(merged_env) \
                .with_args(argument_group.args()) \
                .with_user_overridden_env(
                    alias.get_user_overridden_envs() + resolved_task.get_user_overridden_envs()
                )

            resolved_tasks.append(new_task)

        return GroupDeclaration(name, resolved_tasks, alias.get_description())


class ContextFactory:
    """
    Takes responsibility of loading all tasks defined in USER PROJECT, USER HOME and GLOBALLY
    """

    def __init__(self, io: SystemIO):
        self._io = io

    def _load_context_from_directory(self, path: str) -> ApplicationContext:
        if not os.path.isdir(path):
            raise Exception('Path "%s" not found' % path)

        ctx = ApplicationContext([], [], path)
        contexts = []

        if os.path.isfile(path + '/makefile.py'):
            contexts.append(self._load_from_py(path))

        if os.path.isfile(path + '/makefile.yaml'):
            contexts.append(self._load_from_yaml(path, 'makefile.yaml'))

        if os.path.isfile(path + '/makefile.yml'):
            contexts.append(self._load_from_yaml(path, 'makefile.yml'))

        if not contexts:
            raise ContextFileNotFoundException(path)

        for subctx in contexts:
            ctx = ApplicationContext.merge(ctx, subctx)

        return ctx

    def _load_from_yaml(self, path: str, filename: str) -> ApplicationContext:
        makefile_path = path + '/' + filename

        with open(makefile_path, 'rb') as handle:
            imported, tasks = YamlSyntaxInterpreter(self._io, YamlFileLoader([])).parse(
                handle.read().decode('utf-8'), path, makefile_path
            )

            # Issue 33: Support mixed declarations in imports()
            imports, aliases = distinct_imports(makefile_path, imported)

            return ApplicationContext(tasks=imports, aliases=tasks + aliases, directory=path)

    @staticmethod
    def _load_from_py(path: str):
        makefile_path = path + '/makefile.py'

        if not os.path.isfile(makefile_path):
            raise PythonContextFileNotFoundException(makefile_path)

        try:
            sys.path.append(path)
            makefile = SourceFileLoader("Makefile", makefile_path).load_module()

        except ImportError as e:
            print_exc()
            raise NotImportedClassException(e)

        # Issue 33: Support mixed declarations in imports()
        imports, aliases = distinct_imports(makefile_path, makefile.IMPORTS if "IMPORTS" in dir(makefile) else [])

        return ApplicationContext(
            tasks=imports,
            aliases=(makefile.TASKS if "TASKS" in dir(makefile) else []) + aliases,
            directory=path
        )

    def create_unified_context(self, chdir: str = '') -> ApplicationContext:
        """
        Creates a merged context in order:
        - Internal/Core (this package)
        - System-wide (/usr/lib/rkd)
        - User-home ~/.rkd
        - Application (current directory ./.rkd)
        :return:
        """

        paths = [
            CURRENT_SCRIPT_PATH + '/internal',
            '/usr/lib/rkd',
            os.path.expanduser('~/.rkd'),
            os.getcwd() + '/.rkd'
        ]

        if chdir:
            paths += chdir + '/.rkd'

        if os.getenv('RKD_PATH'):
            paths += os.getenv('RKD_PATH', '').split(':')

        # export for usage inside in makefiles
        os.environ['RKD_PATH'] = ":".join(paths)

        ctx = ApplicationContext([], [], '')

        for path in paths:
            # not all paths could exist, we consider this, we look where it is possible
            if not os.path.isdir(path):
                continue

            try:
                ctx = ApplicationContext.merge(ctx, self._load_context_from_directory(path))
            except ContextFileNotFoundException:
                pass

        ctx.compile()
        ctx.io = self._io

        return ctx


def distinct_imports(file_path: str, imported: List[Union[TaskDeclaration, TaskAliasDeclaration]]) \
        -> Tuple[List[TaskDeclaration], List[TaskAliasDeclaration]]:

    """Separates TaskDeclaration and TaskAliasDeclaration into separate lists, doing validation by the way

    :url: https://github.com/riotkit-org/riotkit-do/issues/33
    """

    aliases = []
    imports = []

    for declaration in imported:
        if isinstance(declaration, TaskAliasDeclaration):
            aliases.append(declaration)
        elif isinstance(declaration, TaskDeclaration):
            imports.append(declaration)
        else:
            raise ContextException('"imports()" from "%s" contains invalid type declaration "%s"' % (
                file_path, str(declaration)
            ))

    return imports, aliases



#!/usr/bin/env python3

import unittest
from rkd.standardlib import InitTask
from rkd.contract import ExecutionContext
from rkd.contract import ArgumentEnv
from rkd.syntax import TaskDeclaration
from rkd.exception import MissingInputException
from rkd.api.inputoutput import IO


class TestExecutionContext(unittest.TestCase):
    def test_get_arg_or_env(self):
        """Checks logic of fetching commandline switch, with fallback to environment variable
        """

        datasets = {
            'env, switch = switch': {
                'switches': {'revolution': 'yes'},
                'envs': {'REVOLUTION': 'no'},
                'defined_args': {
                    '--revolution': {'default': None}
                },
                'declared_envs': {
                    'REVOLUTION': None
                },
                'expects': 'yes',
                'raises': None,
                'test_switch': '--revolution'
            },

            'ENV - NO switch = env': {
                'switches': {'revolution': None},
                'envs': {'REVOLUTION': 'yup'},
                'defined_args': {
                    '--revolution': {'default': None}
                },
                'declared_envs': {
                    'REVOLUTION': None
                },
                'expects': 'yup',
                'raises': None,
                'test_switch': '--revolution'
            },

            'NO env, NO switch = raise': {
                'switches': {'revolution': None},
                'envs': {},
                'defined_args': {
                    '--revolution': {'default': None}
                },
                'declared_envs': {
                    'REVOLUTION': None
                },
                'expects': '',
                'raises': MissingInputException,
                'test_switch': '--revolution'
            },

            'NO env, switch = switch': {
                'switches': {'revolution': 'yes'},
                'envs': {},
                'defined_args': {
                    '--revolution': {'default': None}
                },
                'declared_envs': {
                    'REVOLUTION': None
                },
                'expects': 'yes',
                'raises': None,
                'test_switch': '--revolution'
            },

            'ENV present, SWITCH is set to DEFAULT VALUE = env': {
                'switches': {'person': 'Nobody'},
                'defined_args': {
                    '--person': {'default': 'Nobody'}
                },
                'envs': {'PERSON': 'Gaetano Bresci'},
                'declared_envs': {
                    'PERSON': None
                },
                'expects': 'Gaetano Bresci',
                'raises': False,
                'test_switch': '--person'
            },

            'Custom ENV present, should map a switch to custom ENV name': {
                'switches': {'person': 'Nobody'},
                'defined_args': {
                    '--person': {'default': 'Nobody'}
                },
                'envs': {'PERSON_NAME': 'Luigi Lucheni'},
                'declared_envs': {
                    'PERSON_NAME': ArgumentEnv(name='PERSON_NAME', switch='--person', default='Nobody')
                },
                'expects': 'Luigi Lucheni',
                'raises': False,
                'test_switch': '--person'
            }
        }

        #
        # Actual test code
        #
        def test(dataset_name, dataset):
            task = InitTask()
            task._io = IO()
            task.get_declared_envs = lambda: dataset['declared_envs']

            context = ExecutionContext(TaskDeclaration(task), args=dataset['switches'], env=dataset['envs'],
                                       defined_args=dataset['defined_args'])
            self.assertEqual(dataset['expects'], context.get_arg_or_env(dataset['test_switch']),
                             msg='Dataset failed: %s' % dataset_name)

        #
        # Iteration over datasets
        #
        for dataset_name, dataset in datasets.items():
            with self.subTest(dataset_name):
                if dataset['raises']:
                    with self.assertRaises(dataset['raises']):
                        test(dataset_name, dataset)
                else:
                    test(dataset_name, dataset)





from jsonschema import ValidationError


class ContextException(Exception):
    pass


class TaskNotFoundException(ContextException):
    pass


class InterruptExecution(Exception):
    pass


class TaskException(ContextException):
    pass


class UndefinedEnvironmentVariableUsageError(TaskException):
    pass


class EnvironmentVariableNotUsed(TaskException):
    pass


class EnvironmentVariableNameNotAllowed(TaskException):
    def __init__(self, var_name: str):
        super().__init__('Environment variable with this name "' + var_name + '" cannot be declared, it probably a' +
                         ' commonly reserved name by operating systems')


class UserInputException(Exception):
    pass


class NotSupportedEnvVariableError(UserInputException):
    pass


class YamlParsingException(ContextException):
    """Logic or syntax errors in makefile.yaml"""


class YAMLFileValidationError(YamlParsingException):
    """Errors related to schema validation"""

    def __init__(self, err: ValidationError):
        super().__init__('YAML schema validation failed at path "%s" with error: %s' % (
            '.'.join(err.path), err.message
        ))


class DeclarationException(ContextException):
    """Something wrong with the makefile.py/makefile.yaml """


class ContextFileNotFoundException(ContextException):
    """When makefile.py, makefile.yaml, makefile.yml not found (at least one needed)"""

    def __init__(self, path: str):
        super().__init__('The directory "%s" should contain at least makefile.py, makefile.yaml or makefile.yml' % path)


class PythonContextFileNotFoundException(ContextFileNotFoundException):
    """When makefile.py is not found"""

    def __init__(self, path: str):
        super().__init__('Python context file not found at "%s"' % path)


class NotImportedClassException(ContextException):
    """When class was not imported"""

    def __init__(self, exc: ImportError):
        super().__init__(
            'Your Makefile contains a reference to not available or not installed Python module "%s"' % str(exc)
        )


class EnvironmentVariablesFileNotFound(ContextException):
    """.env file specified, but not existing"""

    def __init__(self, path: str, lookup_paths: list):
        super().__init__(
            'Specified file "%s" as environment variable provider does not exist. Looked in: %s' % (
                path, str(lookup_paths)
            )
        )


class RuntimeException(Exception):
    pass


class MissingInputException(RuntimeException):
    def __init__(self, arg_name: str, env_name: str):
        super().__init__('Either "%s" switch not used, either "%s" was not defined in environment' % (
            arg_name, env_name
        ))




"""
SYNTAX (part of API)
====================

Classes used in a declaration syntax in makefile.py

"""

import os
from typing import List, Dict
from copy import deepcopy
from .contract import TaskDeclarationInterface
from .contract import GroupDeclarationInterface
from .contract import TaskInterface
from ..exception import DeclarationException


class TaskDeclaration(TaskDeclarationInterface):
    _task: TaskInterface
    _env: Dict[str, str]       # environment at all
    _user_defined_env: list    # list of env variables overridden by user
    _args: List[str]

    def __init__(self, task: TaskInterface, env: Dict[str, str] = {}, args: List[str] = []):
        if not isinstance(task, TaskInterface):
            raise DeclarationException('Invalid class: TaskDeclaration needs to take TaskInterface as task argument')

        self._task = task
        self._env = merge_env(env)
        self._args = args
        self._user_defined_env = list(env.keys())

    def to_full_name(self):
        return self._task.get_full_name()

    def with_env(self, envs: Dict[str, str]):
        """ Immutable environment setter. Produces new object each time. """

        copy = deepcopy(self)
        copy._env = envs

        return copy

    def with_args(self, args: List[str]):
        """ Immutable arguments setter. Produces new object each time """

        copy = deepcopy(self)
        copy._args = args

        return copy

    def with_user_overridden_env(self, env_list: list):
        """ Immutable arguments setter. Produces new object each time """

        copy = deepcopy(self)
        copy._user_defined_env = env_list

        return copy

    def get_args(self) -> List[str]:
        return self._args

    def get_task_to_execute(self) -> TaskInterface:
        return self._task

    def to_list(self) -> list:
        return [self]

    def get_env(self):
        return self._env

    def get_user_overridden_envs(self) -> list:
        """ Lists environment variables which were overridden by user """

        return self._user_defined_env

    def get_group_name(self) -> str:
        split = self.to_full_name().split(':')
        return split[1] if len(split) >= 3 else ''

    def get_task_name(self) -> str:
        split = self.to_full_name().split(':')

        if len(split) >= 3:
            return split[2]

        try:
            return split[1]
        except KeyError:
            return self.to_full_name()

    def get_description(self) -> str:
        task = self.get_task_to_execute()

        if task.get_description():
            return task.get_description()

        return task.__doc__.strip().split("\n")[0] if task.__doc__ else ''

    def get_full_description(self) -> str:
        task = self.get_task_to_execute()

        if task.get_description():
            return task.get_description()

        return task.__doc__.strip() if task.__doc__ else ''

    @staticmethod
    def parse_name(name: str) -> tuple:
        split = name.split(':')
        task_name = ":" + split[-1]
        group = ":".join(split[:-1])

        return task_name, group

    def format_task_name(self, name: str) -> str:
        return self.get_task_to_execute().format_task_name(name)

    def __str__(self):
        return 'TaskDeclaration<%s>' % self.get_task_to_execute().get_full_name()


class GroupDeclaration(GroupDeclarationInterface):
    """ Internal DTO: Processed definition of TaskAliasDeclaration into TaskDeclaration """

    _name: str
    _declarations: List[TaskDeclaration]
    _description: str

    def __init__(self, name: str, declarations: List[TaskDeclaration], description: str):
        self._name = name
        self._declarations = declarations
        self._description = description

    def get_declarations(self) -> List[TaskDeclaration]:
        return self._declarations

    def get_name(self) -> str:
        return self._name

    def get_group_name(self) -> str:
        split = self._name.split(':')
        return split[1] if len(split) >= 3 else ''

    def get_task_name(self) -> str:
        split = self._name.split(':')

        if len(split) >= 3:
            return split[2]

        try:
            return split[1]
        except KeyError:
            return self._name

    def to_full_name(self):
        return self.get_name()

    def get_description(self) -> str:
        return self._description

    def format_task_name(self, name: str) -> str:
        return name


class TaskAliasDeclaration:
    """ Allows to define a custom task name that triggers other tasks in proper order """

    _name: str
    _arguments: List[str]
    _env: Dict[str, str]
    _user_defined_env: list  # list of env variables overridden by user
    _description: str

    def __init__(self, name: str, to_execute: List[str], env: Dict[str, str] = {}, description: str = ''):
        self._name = name
        self._arguments = to_execute
        self._env = merge_env(env)
        self._user_defined_env = list(env.keys())
        self._description = description

    def get_name(self):
        return self._name

    def get_arguments(self) -> List[str]:
        return self._arguments

    def get_env(self) -> Dict[str, str]:
        return self._env

    def get_user_overridden_envs(self) -> list:
        """ Lists environment variables which were overridden by user """

        return self._user_defined_env

    def get_description(self) -> str:
        return self._description


def merge_env(env: Dict[str, str]):
    """Merge custom environment variables set per-task with system environment
    """

    # todo: Unit test
    merged_dict = deepcopy(env)
    merged_dict.update(dict(os.environ))

    return merged_dict



#!/usr/bin/env python3

import unittest
import subprocess


class PublishTaskTest(unittest.TestCase):
    def _call_publish_via_shell(self, arguments: str):
        try:
            out = subprocess.check_output('''
                export RKD_PATH=$(pwd)/../example/.rkd
                
                cd ../
                python3 -m rkd --silent :sh -c "python3 -m rkd --silent :py:publish ''' + arguments + '''"
            ''', shell=True)
        except subprocess.CalledProcessError as e:
            return e.output

        return out

    def test_integration_is_output_properly_catch_by_sh(self):
        """Case: In the past there were problems with output displaying by RKD-in-RKD - sh('rkd :py:publish ...')

        Integration test with TaskUtilities.sh()
        """

        out = self._call_publish_via_shell('--username=wrong --password=wrong --test').decode('utf-8')
        self.assertIn('403 Invalid or non-existent authentication information.', out)
        self.assertIn('Uploading distributions to https://test.pypi.org', out)




from typing import Union
from collections import OrderedDict
from ..api.syntax import TaskDeclaration
from ..api.syntax import GroupDeclaration
from ..inputoutput import SystemIO


STATUS_STARTED = 'started'
STATUS_ERRORED = 'errored'
STATUS_FAILURE = 'failure'
STATUS_SUCCEED = 'succeed'


class QueueItem(object):
    task: TaskDeclaration
    status: str

    def __init__(self, task: TaskDeclaration, status: str):
        self.task = task
        self.status = status


class ProgressObserver(object):
    _io: SystemIO
    _tasks: OrderedDict  # OrderedDict[str, QueueItem]
    _failed_count: int

    def __init__(self, io: SystemIO):
        self._io = io
        self._tasks = OrderedDict()
        self._failed_count = False

    @staticmethod
    def _format_parent_task(parent: Union[GroupDeclaration, None]) -> str:
        return ('[part of ' + parent.get_name() + ']') if parent else ''

    def task_started(self, declaration: TaskDeclaration, parent: Union[GroupDeclaration, None], args: list):
        """ When task is just started """

        self._tasks[declaration.to_full_name()] = QueueItem(declaration, STATUS_STARTED)

        self._io.info_msg(' >> Executing %s %s %s' % (
            declaration.to_full_name(),
            ' '.join(args),
            self._format_parent_task(parent)
        ))

    def task_errored(self, declaration: TaskDeclaration, exception: Exception):
        """ On exception catched in task execution """

        self._tasks[declaration.to_full_name()] = QueueItem(declaration, STATUS_ERRORED)
        self._failed_count += 1

        self._io.print_opt_line()
        self._io.error_msg('The task "%s" was interrupted with an %s' % (
            declaration.to_full_name(),
            str(exception.__class__)
        ))
        self._io.print_separator()
        self._io.print_opt_line()

    def task_failed(self, declaration: TaskDeclaration, parent: Union[GroupDeclaration, None]):
        """ When task returns False """

        self._set_status(declaration, STATUS_FAILURE)
        self._failed_count += 1

        if not declaration.get_task_to_execute().is_silent_in_observer():
            self._io.print_opt_line()
            self._io.error_msg('The task "%s" %s ended with a failure' % (
                declaration.to_full_name(),
                self._format_parent_task(parent)
            ))
            self._io.print_separator()
            self._io.print_opt_line()

    def task_succeed(self, declaration: TaskDeclaration, parent: Union[GroupDeclaration, None]):
        """ When task success """

        self._set_status(declaration, STATUS_SUCCEED)

        if not declaration.get_task_to_execute().is_silent_in_observer():
            self._io.print_opt_line()
            self._io.success_msg('The task "%s" %s succeed.' % (
                declaration.to_full_name(),
                self._format_parent_task(parent)
            ))
            self._io.print_separator()
            self._io.print_opt_line()

    def execution_finished(self):
        """ When all tasks were executed """

        if self.has_at_least_one_failed_task():
            self._io.error_msg('Execution failed with %i failed tasks of %i total tasks scheduled for execution' % (
                self._failed_count, len(self._tasks)
            ))
        else:
            self._io.success_msg('Successfully executed %i tasks.' % len(self._tasks))

        self._io.print_opt_line()

    def _set_status(self, declaration: TaskDeclaration, status: str):
        self._tasks[declaration.to_full_name()] = QueueItem(declaration, status)

    def has_at_least_one_failed_task(self) -> bool:
        return self._failed_count > 0



"""
Forked executor takes a Python code and executes as a separate process.

Schema:
   1. RKD spawns PYTHON process that executes THIS SCRIPT
   2. RKD creates a temporary file for communication
   3. Context and task code is passed to communication file, so the 'forked' subprocess can read it
   4. Process is unpacking the serialized data, executing the task
   5. Possible errors are catch and passed via temporary file to RKD (to re-raise in main process)
   6. Return result is passed via temporary file to RKD

This method gives us a possibility to change user on-the-fly (in Docker containers it is very helpful) and possibly
a chance in the future to implement eg. remote executor or at least partially isolated executor
"""

import pickle


FORKED_EXECUTOR_TEMPLATE = """
import pickle
import base64
import sys

def _communicate_return(val):
    with open(communication_file, 'wb') as f:
        f.write(pickle.dumps(val))


communication_file = sys.stdin.read().strip()

try:
    #
    # Load serialized context
    #
    
    with open(communication_file, 'rb') as f:
        unserialized = pickle.loads(f.read())
    
    task = unserialized['task']
    ctx = unserialized['ctx']

    if not task.execute(ctx):
        _communicate_return(False)
        sys.exit(0)

except Exception as exc:
    _communicate_return(exc)

    sys.exit(0)

_communicate_return(True)
"""


def get_unpicklable(instance, exception=None, string='', first_only=True):
    """
    Recursively go through all attributes of instance and return a list of whatever
    can't be pickled.

    Set first_only to only print the first problematic element in a list, tuple or
    dict (otherwise there could be lots of duplication).

    See: https://stackoverflow.com/a/55224405/6782994 (author)
    """
    problems = []
    if isinstance(instance, tuple) or isinstance(instance, list):
        for k, v in enumerate(instance):
            try:
                pickle.dumps(v)
            except BaseException as e:
                problems.extend(get_unpicklable(v, e, string + f'[{k}]'))
                if first_only:
                    break
    elif isinstance(instance, dict):
        for k in instance:
            try:
                pickle.dumps(k)
            except BaseException as e:
                problems.extend(get_unpicklable(
                    k, e, string + f'[key type={type(k).__name__}]'
                ))
                if first_only:
                    break
        for v in instance.values():
            try:
                pickle.dumps(v)
            except BaseException as e:
                problems.extend(get_unpicklable(
                    v, e, string + f'[val type={type(v).__name__}]'
                ))
                if first_only:
                    break
    else:
        for k, v in instance.__dict__.items():
            try:
                pickle.dumps(v)
            except BaseException as e:
                problems.extend(get_unpicklable(v, e, string + '.' + k))

    # if we get here, it means pickling instance caused an exception (string is not
    # empty), yet no member was a problem (problems is empty), thus instance itself
    # is the problem.
    if string != '' and not problems:
        problems.append(
            string + f" (Type '{type(instance).__name__}' caused: {exception})"
        )

    return problems



#!/usr/bin/env python3

import unittest.mock
from rkd.contract import ArgparseArgument


class TestArgparseArgument(unittest.TestCase):
    def test_arguments_are_registered(self):
        arg = ArgparseArgument(['--name', '-n'], {'default': 'Michail Bakunin'})

        self.assertEqual(['--name', '-n'], arg.args)
        self.assertEqual({'default': 'Michail Bakunin'}, arg.kwargs)



#!/usr/bin/env python3

from setuptools import setup, find_packages

setup(
    setup_requires=['pbr'],
    pbr=True,
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    include_package_data=True
)



#!/usr/bin/env python3

import unittest
from tempfile import NamedTemporaryFile
from rkd.standardlib import LineInFileTask
from rkd.test import mock_task, mock_execution_context
from rkd.api.inputoutput import BufferedSystemIO


class LineInFileTaskTest(unittest.TestCase):
    @staticmethod
    def _execute_mocked_task(params: dict, envs: dict = {}) -> BufferedSystemIO:
        io = BufferedSystemIO()

        task: LineInFileTask = LineInFileTask()
        mock_task(task, io=io)
        task.execute(mock_execution_context(task, params, envs))

        return io

    def test_simply_appends_line_if_it_does_not_exists(self):
        with NamedTemporaryFile() as tmp_file:
            self._execute_mocked_task({
                'file': tmp_file.name,
                '--output': '',
                '--regexp': 'Linux ([0-9.]+)',
                '--insert': 'Linux 4.16.1',
                '--fail-on-no-occurrence': False,
                '--only-first-occurrence': False,
                '--new-after-line': ''
            })

            self.assertEqual("Linux 4.16.1\n", tmp_file.read().decode('utf-8'))

    def test_simply_replaces_found_one_matched_line(self):
        with NamedTemporaryFile() as tmp_file:
            # at first add initial file content
            with open(tmp_file.name, 'w') as fw:
                fw.write('Text: Hello')

            # then modify it
            self._execute_mocked_task({
                'file': tmp_file.name,
                '--output': '',
                '--regexp': 'Text: (.*)',
                '--insert': 'Text: 14 June 1872 unions were legalised in Canada through the Trade Unions Act after a' +
                            ' wave of illegal strikes and protests. Still the law didn\'t compel employers to' +
                            ' recognise or bargain with unions, and picketing remained illegal.',
                '--fail-on-no-occurrence': False,
                '--only-first-occurrence': False,
                '--new-after-line': ''
            })

            self.assertIn("14 June 1872 unions were legalised in Canada", tmp_file.read().decode('utf-8'))

    def test_advanced_replaces_using_regexp_pattern(self):
        """Tests usage of regexp groups eg. $match[0], $match[1] etc."""

        with NamedTemporaryFile() as tmp_file:
            # at first add initial file content
            with open(tmp_file.name, 'w') as fw:
                fw.write('Symbol: 1')

            # then modify it
            self._execute_mocked_task({
                'file': tmp_file.name,
                '--output': '',
                '--regexp': 'Symbol: ([0-9]+)',
                '--insert': 'Symbol: $match[0] + 161',
                '--fail-on-no-occurrence': False,
                '--only-first-occurrence': False,
                '--new-after-line': ''
            })

            self.assertIn("Symbol: 1 + 161", tmp_file.read().decode('utf-8'))

    def test_fails_when_no_occurrence_and_flag_is_set(self):
        """Test flag --fail-on-no-occurrence"""

        with NamedTemporaryFile() as tmp_file:
            io = self._execute_mocked_task({
                'file': tmp_file.name,
                '--output': '',
                '--regexp': 'Knock knock',
                '--insert': 'Who\'s there?',
                '--fail-on-no-occurrence': True,
                '--only-first-occurrence': False,
                '--new-after-line': ''
            })

            self.assertIn('No matching line for selected regexp found', io.get_value())

    def test_replaces_only_first_occurrence(self):
        """Test that only one occurrence will be replaced"""

        with NamedTemporaryFile() as tmp_file:
            # at first add initial file content
            with open(tmp_file.name, 'w') as fw:
                fw.write("Symbol: 1\n")
                fw.write("Color: Red\n")
                fw.write("Symbol: 2\n")
                fw.write("\n\n")
                fw.write("Dog or cat: Dog\n")

            # then modify it
            self._execute_mocked_task({
                'file': tmp_file.name,
                '--output': '',
                '--regexp': 'Symbol: ([0-9]+)',
                '--insert': 'Symbol: 161',
                '--fail-on-no-occurrence': True,
                '--only-first-occurrence': True,
                '--new-after-line': ''
            })

            processed_file_content = tmp_file.read().decode('utf-8')

            self.assertIn("Symbol: 161", processed_file_content)
            self.assertNotIn("Symbol: 1\n", processed_file_content)
            self.assertIn("Symbol: 2", processed_file_content)

    def test_adds_multiple_lines_after_selected_marker(self):
        self.maxDiff = None

        task = LineInFileTask()
        task._io = BufferedSystemIO()

        result = task._insert_new_lines_if_necessary(
            False,
            '''
            Hijo del pueblo, te oprimen cadenas,
            y esa injusticia no puede seguir;
            si tu existencia es un mundo de penas
            antes que esclavo prefiere morir.
            En la batalla, la hiena fascista.
            por nuestro esfuerzo sucumbirá;
            y el pueblo entero, con los anarquistas,
            hará que triunfe la libertad.
            
            --- Read more --
            
            --- EOF
            ''',
            lines_to_insert='''            Trabajador, no más sufrir,
            el opresor ha de sucumbir.
            Levántate, pueblo leal,
            al grito de revolución social.
            Fuerte unidad de fe y de acción
            producirá la revolución.
            Nuestro pendón uno ha de ser:
            sólo en la unión está el vencer.'''.split("\n"),
            after_line_regexp='(.*)Read\ more(.*)',
            only_first_occurrence=True,
            regexp='.*Trabajador.*'
        )

        self.assertEqual(
            '''
            Hijo del pueblo, te oprimen cadenas,
            y esa injusticia no puede seguir;
            si tu existencia es un mundo de penas
            antes que esclavo prefiere morir.
            En la batalla, la hiena fascista.
            por nuestro esfuerzo sucumbirá;
            y el pueblo entero, con los anarquistas,
            hará que triunfe la libertad.
            
            --- Read more --
            Trabajador, no más sufrir,
            el opresor ha de sucumbir.
            Levántate, pueblo leal,
            al grito de revolución social.
            Fuerte unidad de fe y de acción
            producirá la revolución.
            Nuestro pendón uno ha de ser:
            sólo en la unión está el vencer.
            
            --- EOF
            ''',
            result,
        )

    def test_adds_missing_lines_after_specified_markers(self):
        task = LineInFileTask()
        task._io = BufferedSystemIO()

        result = task._insert_new_lines_if_necessary(
            False,
            '''
            Press list
            
            [@CNT]
            Other press
            Solidaridad Obrera
            
            [@FAI]
            Tierra y Libertad
            ''',
            lines_to_insert=['            Other press'],
            after_line_regexp='.*(CNT|FAI).*',
            only_first_occurrence=False,
            regexp='.*Other press'
        )

        self.assertEqual(
            '''
            Press list
            
            [@CNT]
            Other press
            Solidaridad Obrera
            
            [@FAI]
            Other press
            Tierra y Libertad
            ''',
            result
        )

    def test_lines_are_added_only_once_even_when_command_called_multiple_times(self):
        """Check that multiple invocations wont duplicate the line"""

        with NamedTemporaryFile() as tmp_file:
            for i in range(0, 10):
                self._execute_mocked_task({
                    'file': tmp_file.name,
                    '--output': '',
                    '--regexp': 'Bakunin ([0-9.]+)',
                    '--insert': 'Bakunin 5.16.%i' % i,
                    '--fail-on-no-occurrence': False,
                    '--only-first-occurrence': False,
                    '--new-after-line': ''
                })

            # "9" is because we have 9 iterations in range
            self.assertEqual("Bakunin 5.16.9\n", tmp_file.read().decode('utf-8'))


import os
import re
from typing import Pattern
from argparse import ArgumentParser
from subprocess import CalledProcessError
from jinja2 import Environment
from jinja2 import FileSystemLoader
from jinja2 import StrictUndefined
from jinja2.exceptions import UndefinedError
from ..api.contract import TaskInterface
from ..api.contract import ExecutionContext
from ..api.syntax import TaskDeclaration


class FileRendererTask(TaskInterface):
    """Renders a .j2 file using environment as input variables"""

    def get_name(self) -> str:
        return ':render'

    def get_group_name(self) -> str:
        return ':j2'

    def execute(self, context: ExecutionContext) -> bool:
        source = context.get_arg('--source')
        output = context.get_arg('--output')

        if not os.path.isfile(source):
            self.io().error_msg('Source file does not exist at path "%s"' % source)
            return False

        if output != '-' and not os.path.isdir(os.path.dirname(output)):
            self.sh('mkdir -p %s' % os.path.dirname(output))

        with open(source, 'rb') as f:
            # @todo: Support for .rkd directories in proper order
            tpl = Environment(loader=FileSystemLoader(['./', './.rkd/templates']), undefined=StrictUndefined)\
                .from_string(f.read().decode('utf-8'))

            try:
                rendered = tpl.render(**context.env)
            except UndefinedError as e:
                self.io().error_msg('Undefined variable - ' + str(e))
                return False

            if output == "-":
                self.io().outln(rendered)
            else:
                with open(output, 'wb') as t:
                    t.write(rendered.encode('utf-8'))

            return True

    def configure_argparse(self, parser: ArgumentParser):
        parser.description = 'Renders a JINJA2 file. Environment variables are accessible in templates.'
        parser.add_argument('--source', '-s', help='Template file', required=True)
        parser.add_argument('--output', '-o', help='Output to file, set "-" for stdout', default='-')


class RenderDirectoryTask(TaskInterface):
    """Renders *.j2 files recursively in a directory to other directory"""

    def get_name(self) -> str:
        return ':directory-to-directory'

    def get_group_name(self) -> str:
        return ':j2'

    def execute(self, context: ExecutionContext) -> bool:
        source_root = context.get_arg('--source')
        target_root = context.get_arg('--target')
        delete_source_files = context.get_arg('--delete-source-files')
        pattern = re.compile(context.get_arg('--pattern'))
        exclude_pattern = re.compile(context.get_arg('--exclude-pattern')) if context.get_arg('--exclude-pattern') else None
        copy_not_matched = context.get_arg('--copy-not-matching-files')
        template_filenames = context.get_arg('--template-filenames')

        self.io().info_msg('Pattern is `%s`' % context.get_arg('--pattern'))

        for root, subdirs, files in os.walk(source_root):
            for file in files:
                source_full_path = root + '/' + file
                target_full_path = target_root + '/' + source_full_path[len(source_root):]

                if target_full_path.endswith('.j2'):
                    target_full_path = target_full_path[:-3]

                if template_filenames:
                    target_full_path = self.replace_vars_in_filename(context.env, target_full_path)

                if exclude_pattern and self._is_file_matching_filter(exclude_pattern, source_full_path):
                    self.io().info_msg('Skipping file "%s" - (filtered out by --exclude-pattern)' % source_full_path)
                    continue

                if not self._is_file_matching_filter(pattern, source_full_path):
                    if copy_not_matched:
                        self.io().info_msg('Copying "%s" regular file' % source_full_path)
                        self._copy_file(source_full_path, target_full_path)

                        continue

                    self.io().info_msg('Skipping file "%s" (filtered out by --pattern)' % source_full_path)
                    continue

                self.io().info_msg('Rendering file "%s" into "%s"' % (source_full_path, target_full_path))

                if not self._render(source_full_path, target_full_path):
                    # stderr will be passed through
                    return False

                if delete_source_files:
                    self._delete_file(source_full_path)

        return True

    @staticmethod
    def replace_vars_in_filename(env_vars: dict, filename: str) -> str:
        for name, value in env_vars.items():
            filename = filename.replace('--%s--' % name, value)

        return filename

    def _copy_file(self, source_full_path: str, target_full_path: str):
        self.sh('mkdir -p "%s"' % os.path.dirname(target_full_path))
        self.sh('cp -p "%s" "%s"' % (source_full_path, target_full_path))

    def _render(self, source_path: str, target_path: str) -> bool:
        try:
            self.rkd([':j2:render', '--source="%s"' % source_path, '--output="%s"' % target_path], verbose=True)
            return True

        except CalledProcessError:
            return False

    @staticmethod
    def _delete_file(full_path: str):
        os.unlink(full_path)

    @staticmethod
    def _is_file_matching_filter(pattern: Pattern, full_path: str):
        return pattern.match(full_path) is not None

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--source', '-s', help='Source path where templates are stored', required=True)
        parser.add_argument('--target', '-t', help='Target path where templates should be rendered', required=True)
        parser.add_argument('--delete-source-files', '-d', help='Delete source files after rendering?', default=False,
                            action='store_true')
        parser.add_argument('--pattern', '-p', help='Optional regexp pattern to match full paths',
                            default='(.*).j2')
        parser.add_argument('--exclude-pattern', '-xp', help='Optional regexp for a pattern exclude, to exclude files')
        parser.add_argument('--copy-not-matching-files', '-c', help='Copy all files that are not matching the pattern' +
                                                                    ' instead of skipping them', action='store_true')
        parser.add_argument('--template-filenames', '-tf',
                            help='Replace variables in filename eg. --VAR--, ' +
                                 'where VAR is a name of environment variable', action='store_true')


def imports():
    return [
        TaskDeclaration(FileRendererTask()),
        TaskDeclaration(RenderDirectoryTask())
    ]



#!/usr/bin/env python3

import unittest
import os
from io import StringIO
from rkd.contract import ExecutionContext
from rkd.api.inputoutput import BufferedSystemIO
from rkd.api.inputoutput import IO
from rkd.test import get_test_declaration
from rkd.test import mock_task
from rkd.execution.declarative import DeclarativeExecutor
from rkd.execution.declarative import Step

CURRENT_SCRIPT_PATH = os.path.dirname(os.path.realpath(__file__))


class TestDeclarativeExecutor(unittest.TestCase):
    @staticmethod
    def _create_callable_tester(code: str, language: str, io: IO = None) -> bool:
        if not io:
            io = IO()

        executor = DeclarativeExecutor()
        declaration = get_test_declaration()
        declaration.get_task_to_execute()._io = io
        ctx = ExecutionContext(declaration)

        step = Step(
            language=language,
            task_name=':test',
            code=code,
            envs={},
            task_num=1,
            rkd_path=''
        )

        if language == 'python':
            return executor.execute_python_step(ctx, declaration.get_task_to_execute(), step)
        else:
            return executor.execute_bash_step(ctx, declaration.get_task_to_execute(), step)

    def test_bash_successful_case(self):
        """ Bash callable test: Successful case """

        io = IO()
        out = StringIO()

        with io.capture_descriptors(stream=out, enable_standard_out=False):
            self._create_callable_tester('python --version', language='bash')

        self.assertIn("Python", out.getvalue())
        self.assertTrue(out.getvalue(), msg='python --version should result with a True')

    def test_bash_failure_case_on_invalid_exit_code(self):
        """ Bash callable test: Check if failures are correctly catched """

        result = self._create_callable_tester('exit 161', language='bash', io=BufferedSystemIO())

        self.assertFalse(result)

    def test_python_case_syntax_error(self):
        result = self._create_callable_tester('''impppport os''', language='python', io=BufferedSystemIO())

        self.assertFalse(result)

    def test_python_case_multiline_with_imports_and_call_to_this(self):
        result = self._create_callable_tester('''
import os


# do a very simple verification of class names, if not defined then it would fail
return "ExecutionContext" in str(ctx) and "Task" in str(this)
    ''', language='python')

        self.assertTrue(result)

    def test_bash_case_verify_env_variables_are_present(self):

        io = IO()
        out = StringIO()

        with io.capture_descriptors(stream=out, enable_standard_out=False):
            self._create_callable_tester('echo "Boolean: ${ARG_TEST}, Text: ${ARG_MESSAGE}"', language='bash')

        self.assertIn('ARG_TEST: unbound variable', out.getvalue())

    def test_executing_multiple_steps_one_by_one_the_order_is_preserved(self):
        """Assert that execution order is preserved - as we register steps"""

        io = IO()
        str_io = StringIO()

        task_declaration = get_test_declaration()
        mock_task(task_declaration.get_task_to_execute(), io=io)

        ctx = ExecutionContext(task_declaration)
        executor = DeclarativeExecutor()
        executor.add_step('python', 'this.io().outln("First"); return True', task_name=':first', rkd_path='', envs={})
        executor.add_step('bash', 'echo "Second"; exit 0', task_name=':second', rkd_path='', envs={})
        executor.add_step('python', 'this.io().outln("Third"); return True', task_name=':third', rkd_path='', envs={})

        with io.capture_descriptors(target_files=[], stream=str_io, enable_standard_out=False):
            executor.execute_steps_one_by_one(ctx, task_declaration.get_task_to_execute())

        self.assertEqual("First\nSecond\nThird\n", str_io.getvalue())

    def test_one_failed_step_is_preventing_next_steps_from_execution_and_result_is_marked_as_failure(self):
        """Check the correctness of error handling"""

        io = IO()
        str_io = StringIO()
        buffered = BufferedSystemIO()

        task_declaration = get_test_declaration()
        mock_task(task_declaration.get_task_to_execute(), io=buffered)

        ctx = ExecutionContext(task_declaration)
        executor = DeclarativeExecutor()
        executor.add_step('python', 'this.io().outln("Peter Kropotkin"); return True', task_name=':first', rkd_path='', envs={})
        executor.add_step('bash', 'echo "Buenaventura Durruti"; exit 1', task_name=':second', rkd_path='', envs={})
        executor.add_step('python', 'this.io().outln("This one will not show"); return True', task_name=':third', rkd_path='', envs={})

        with io.capture_descriptors(target_files=[], stream=str_io, enable_standard_out=False):
            final_result = executor.execute_steps_one_by_one(ctx, task_declaration.get_task_to_execute())

        output = str_io.getvalue() + buffered.get_value()

        self.assertIn('Peter Kropotkin', output)
        self.assertIn('Buenaventura Durruti', output)
        self.assertNotIn('This one will not show', output)
        self.assertEqual(False, final_result)

    def test_return_false_is_added_to_the_code(self):
        """Assert that code in Python without a "return" will have "return false" by default

        Previously, when return was not added automatically there was an error raised (not always - depending on the code), example:
            AttributeError: 'Try' object has no attribute 'value'

        https://github.com/riotkit-org/riotkit-do/issues/37
        :return:
        """

        io = IO()
        out = StringIO()

        with io.capture_descriptors(stream=out, enable_standard_out=False):
            returned_value = self._create_callable_tester(
                '''print('History isn't made by kings and politicians, it is made by us.');''', language='python')

        self.assertIn("History isn't made by kings and politicians, it is made by us.", out.getvalue())
        self.assertFalse(returned_value)



#!/usr/bin/env python3

import unittest
import unittest.mock
import os
from tempfile import NamedTemporaryFile
from rkd.context import ContextFactory
from rkd.context import ApplicationContext
from rkd.context import distinct_imports
from rkd.api.inputoutput import NullSystemIO
from rkd.exception import ContextException
from rkd.syntax import TaskDeclaration
from rkd.syntax import TaskAliasDeclaration
from rkd.test import TestTask

CURRENT_SCRIPT_PATH = os.path.dirname(os.path.realpath(__file__))


class ContextTest(unittest.TestCase):
    def test_loads_internal_context(self) -> None:
        """Test if internal context (RKD by default has internal context) is loaded properly
        """
        discovery = ContextFactory(NullSystemIO())
        ctx = discovery._load_context_from_directory(CURRENT_SCRIPT_PATH + '/../src/rkd/internal')

        self.assertTrue(isinstance(ctx, ApplicationContext))

    def test_loads_internal_context_in_unified_context(self) -> None:
        """Check if application loads context including paths from RKD_PATH
        """

        os.environ['RKD_PATH'] = CURRENT_SCRIPT_PATH + '/../docs/examples/makefile-like/.rkd'
        ctx = None

        try:
            discovery = ContextFactory(NullSystemIO())
            ctx = discovery.create_unified_context()
        except:
            raise
        finally:
            self.assertIn(
                ':find-images',
                ctx.find_all_tasks().keys(),
                msg=':find-images is defined in docs/examples/makefile-like/.rkd/makefile.py as an alias type task' +
                    ', expected that it would be loaded from path placed at RKD_PATH'
            )

            os.environ['RKD_PATH'] = ''

    def test_loads_when_only_yaml_file_is_in_directory(self):
        """Check if tasks will be loaded when in .rkd there is only a YAML file
        """

        self._common_test_loads_task_from_file(
            path=CURRENT_SCRIPT_PATH + '/../docs/examples/yaml-only/.rkd',
            task=':hello',
            filename='makefile.yaml'
        )

    def test_loads_when_only_py_file_is_in_directory(self):
        """Check if tasks will be loaded when in .rkd there is only a YAML file
        """

        self._common_test_loads_task_from_file(
            path=CURRENT_SCRIPT_PATH + '/../docs/examples/python-only/.rkd',
            task=':hello-python',
            filename='makefile.py'
        )

    def test_context_remembers_directories_from_which_it_was_loaded(self):
        """Verify that whenever we merge contexts, the 'directories' attribute is also merged"""

        ctx1 = ApplicationContext([], [], '/home/iwa-ait')
        ctx2 = ApplicationContext([], [], '/home/black-lives-matters')

        ctx_merged = ApplicationContext.merge(ctx1, ctx2)

        self.assertEqual(['/home/iwa-ait'], ctx1.directories)
        self.assertEqual(['/home/iwa-ait', '/home/black-lives-matters'], ctx_merged.directories)

    def test_context_empty_path_is_not_applied_to_directories(self):
        """Test that '' path will not be added to directories list"""

        ctx = ApplicationContext([], [], '')

        self.assertEqual([], ctx.directories)
        self.assertEqual(0, len(ctx.directories))

    def test_distint_imports_on_py_file(self):
        """Case: PY file can define a variable IMPORTS that can contain both TaskDeclaration and TaskAliasDeclaration
        """

        with NamedTemporaryFile() as tmp_file:
            tmp_file.write(b'''
            from rkd.syntax import TaskAliasDeclaration as Task
            
            IMPORTS = [
                TaskAliasDeclaration(':hello', [':test'])
            ]
            ''')

            with unittest.mock.patch('rkd.context.os.path.isfile', return_value=True):
                with unittest.mock.patch('rkd.context.SourceFileLoader.load_module') as src_loader_method:
                    class TestImported:
                        IMPORTS = []
                        TASKS = []

                    src_loader_method.return_value = TestImported()
                    src_loader_method.return_value.IMPORTS = [TaskAliasDeclaration(':hello', [':test'])]
                    src_loader_method.return_value.TASKS = []

                    ctx_factory = ContextFactory(NullSystemIO())
                    ctx = ctx_factory._load_from_py(tmp_file.name)

                    self.assertIn(':hello', ctx._task_aliases)

    def test_distinct_imports_on_yaml_file(self):
        """Case: YAML file can import a module that contains imports() method
        And that method returns list of Union[TaskDeclaration, TaskAliasDeclaration]
        """

        with NamedTemporaryFile() as tmp_file:
            tmp_file.write(b'''
            version: org.riotkit.rkd/yaml/v1
            imports:
                - fictional
            ''')

            with unittest.mock.patch('rkd.context.YamlSyntaxInterpreter.parse') as parse_method:
                parse_method.return_value = ([TaskAliasDeclaration(':hello', [':test'])], [])

                ctx_factory = ContextFactory(NullSystemIO())
                ctx = ctx_factory._load_from_yaml(os.path.dirname(tmp_file.name), os.path.basename(tmp_file.name))

                self.assertIn(':hello', ctx._task_aliases)

    def test_distinct_imports_raises_exception_when_unknown_type_object_added_to_list(self):
        self.assertRaises(ContextException,
                          lambda: distinct_imports('hello', ['string-should-not-be-there-even-IDE-knows-that']))

    def test_distinct_imports_does_not_raise_any_exception_when_no_data(self):
        self.assertEqual(([], []), distinct_imports('hello', []))

    def test_distinct_imports_separtes_lists(self):
        """A successful case for distinct_imports()"""

        imports, aliases = distinct_imports('hello', [
            TaskDeclaration(TestTask()),
            TaskAliasDeclaration(':hello', [':test'])
        ])

        self.assertTrue(isinstance(imports[0], TaskDeclaration))
        self.assertTrue(isinstance(aliases[0], TaskAliasDeclaration))

    def _common_test_loads_task_from_file(self, path: str, task: str, filename: str):
        os.environ['RKD_PATH'] = path
        ctx = None

        try:
            discovery = ContextFactory(NullSystemIO())
            ctx = discovery.create_unified_context()
        except:
            raise
        finally:
            self.assertIn(task, ctx.find_all_tasks().keys(),
                          msg='Expected that %s task would be loaded from %s' % (task, filename))

            os.environ['RKD_PATH'] = ''



"""
Generic YAML file parsing module

Uses standard YAML parser, adds additional features, such as:
  - Schema validation
  - RKD lookup paths integration
"""

import os
from typing import List
from yaml import load as yaml_load
from yaml import Loader as YamlLoader
from json import load as json_load
from jsonschema import validate
from jsonschema import ValidationError
from jsonschema import draft7_format_checker
from .exception import YAMLFileValidationError

CURRENT_SCRIPT_PATH = os.path.dirname(os.path.realpath(__file__))


class YamlFileLoader(object):
    """YAML loader extended by schema validation support

    YAML schema is stored as JSON files in .rkd/schema directories.
    The Loader looks in all paths defined in RKD_PATH as well as in paths provided by ApplicationContext
    """

    paths: List[str]

    def __init__(self, paths: List[str]):
        self.paths = paths

    def load_from_file(self, filename: str, schema_name: str):
        """Loads a YAML file from given path, a wrapper to load()"""

        file_path = self.find_path_by_name(filename, '')

        if not file_path:
            raise FileNotFoundError('YAML file "%s" not found in any of lookup paths: %s' % (
                filename, str(self.get_lookup_paths(''))
            ))

        with open(file_path, 'rb') as f:
            return self.load(f.read(), schema_name)

    def load(self, stream, schema_name: str):
        """Loads a YAML, validates and return parsed as dict/list """

        schema_name = schema_name.replace('/', '-') + '.json'
        schema_path = self.find_path_by_name(schema_name, 'schema')

        if not schema_path:
            raise FileNotFoundError('Schema "%s" cannot be found, looked in: %s' % (
                schema_name, str(self.get_lookup_paths('schema'))
            ))

        parsed = yaml_load(stream, YamlLoader)

        with open(schema_path, 'rb') as f:
            try:
                validate(instance=parsed, schema=json_load(f), format_checker=draft7_format_checker)
            except ValidationError as e:
                raise YAMLFileValidationError(e)

        return parsed

    def find_path_by_name(self, filename: str, subdir: str) -> str:
        """Find schema in one of RKD directories or in current path
        """

        for path in self.get_lookup_paths(subdir):
            file_path = path + '/' + filename

            if os.path.isfile(file_path):
                return file_path

        return ''

    def get_lookup_paths(self, subdirectory: str) -> List[str]:
        paths = [os.getcwd(), os.getcwd() + '/' + subdirectory, os.getcwd() + '/.rkd/' + subdirectory]
        global_paths = os.getenv('RKD_PATH', '').split(':')
        global_paths.reverse()

        for path in self.paths:
            paths.append(path + '/' + subdirectory)

        for path in global_paths:
            paths.append(path + '/' + subdirectory)

        paths.append(CURRENT_SCRIPT_PATH + '/internal/' + subdirectory)

        return list(dict.fromkeys(paths))



import os
from abc import ABC
from subprocess import CalledProcessError
from argparse import ArgumentParser
from typing import Dict
from rkd.api.contract import TaskInterface, ExecutionContext
from rkd.api.syntax import TaskDeclaration


class BasePythonTask(TaskInterface, ABC):
    def get_group_name(self) -> str:
        return ':py'


class PublishTask(BasePythonTask):
    """ Publishes Python packages to PIP """

    def get_name(self) -> str:
        return ':publish'

    def get_declared_envs(self) -> Dict[str, str]:
        return {
            'TWINE_PATH': 'twine'
        }

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--username', help='Username in repository', required=True)
        parser.add_argument('--password', help='Password in repository', required=True)
        parser.add_argument('--skip-existing', help='Do not attempt to release twice', action='store_true')
        parser.add_argument('--test', help='Release to a test test.pypi.org', action='store_true')
        parser.add_argument('--url', help='Repository URL', default='')
        parser.add_argument('--src', help='Source files', default='dist/*')

    def execute(self, context: ExecutionContext) -> bool:
        opts = ''

        if context.args['skip_existing']:
            opts += ' --skip-existing '

        if context.get_arg('--url'):
            if context.get_arg('--test'):
                raise Exception('Cannot use --url and --test switch at once')

            opts += ' --repository-url=%s' % context.get_arg('--url')

        if context.get_arg('--test'):
            opts += ' --repository-url https://test.pypi.org/legacy/ '

        self.sh('''
            %s upload \\
                --disable-progress-bar \\
                --verbose \\
                --username=%s \\
                --password=%s \\
                %s %s
        ''' % (
            context.get_env('TWINE_PATH'),
            context.get_arg('--username'),
            context.get_arg('--password'),
            opts,
            context.get_arg('--src')
        ))

        return True


class BuildTask(BasePythonTask):
    """ Builds a Python package in a format to be packaged for publishing """

    def execute(self, context: ExecutionContext) -> bool:
        self.sh('chmod +x setup.py')
        self.sh('./setup.py sdist bdist_wheel %s' % context.get_arg('--args'))

        return True

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--args', '-a', help='Optional arguments', required=False, default='')

    def get_name(self) -> str:
        return ':build'


class InstallTask(BasePythonTask):
    """ Install a Python package using setuptools """

    def execute(self, context: ExecutionContext) -> bool:
        self.sh('chmod +x setup.py')
        self.sh('./setup.py install %s' % context.get_arg('--args'))

        return True

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--args', '-a', help='Optional arguments', required=False, default='')

    def get_name(self) -> str:
        return ':install'


class CleanTask(BasePythonTask):
    """ Clean up the built Python modules """

    def execute(self, context: ExecutionContext) -> bool:
        self._io.info('Cleaning up the built packages')
        self.sh('rm -rf pbr.egg.info .eggs dist build', verbose=True)

        return not os.path.isdir('pbr.egg.info') and not os.path.isdir('.eggs')

    def configure_argparse(self, parser: ArgumentParser):
        pass

    def get_name(self) -> str:
        return ':clean'


class UnitTestTask(BasePythonTask):
    """ Runs unit tests using standard Python framework "unittest" """

    def execute(self, context: ExecutionContext) -> bool:
        cmd = 'export PYTHONUNBUFFERED=1; '

        if context.get_arg('--src-dir'):
            cmd += 'cd %s && ' % context.get_arg('--src-dir')

        cmd += '%s -m unittest discover -s %s ' % (
            context.get_arg('--python-bin'),
            context.get_arg('--tests-dir')
        )

        if context.get_arg('--pattern'):
            cmd += ' -p %s ' % context.get_arg('--pattern')

        if context.args['filter']:
            cmd += ' -k %s ' % context.get_arg('--filter')

        try:
            self.sh(cmd, verbose=True, strict=True)
        except CalledProcessError as e:
            return False

        return True

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--src-dir', default='src', help='Directory where packages are placed')
        parser.add_argument('--tests-dir', default='../test',
                            help='Relative directory to --src-dir where to look for tests')
        parser.add_argument('--pattern', help='Pattern to match tests, default test*.py', default='test*.py')
        parser.add_argument('--filter', '-p', help='Pattern to filter tests')
        parser.add_argument('--python-bin', default='python3', help='Python binary name (if in PATH) or path')

    def get_name(self) -> str:
        return ':unittest'


def imports():
    return [
        TaskDeclaration(CleanTask()),
        TaskDeclaration(PublishTask()),
        TaskDeclaration(BuildTask()),
        TaskDeclaration(InstallTask()),
        TaskDeclaration(UnitTestTask())
    ]



#!/usr/bin/env python3

"""

Test temporary files management
===============================

RKD has a principle to work with a workspace. This rule is best for Continuous Integration systems, where
we often want to work only inside of given directory, never outside.

"""

import os
import unittest
import subprocess
from uuid import uuid4
from rkd.api.temp import TempManager


class TestTempManager(unittest.TestCase):
    oldCwd: str = ''

    def setUp(self) -> None:
        super().setUp()
        self.oldCwd = os.getcwd()
        os.chdir('/tmp')

        try:
            os.mkdir('.rkd')
        except FileExistsError:
            pass

    def tearDown(self) -> None:
        super().tearDown()
        os.chdir(self.oldCwd)

    def test_create_tmp_file_path_assigns_an_unique_filename(self):
        path, filename = TempManager().create_tmp_file_path()

        self.assertFalse(os.path.isfile(path))

    def test_assign_temporary_file_creates_a_file_and_finally_clean_up_removes_file(self):
        manager = TempManager()

        with self.subTest('Creation'):
            path = manager.assign_temporary_file()
            self.assertTrue(os.path.isfile(path))

        with self.subTest('Deletion'):
            manager.finally_clean_up()
            self.assertFalse(os.path.isfile(path))

    def test_assign_temporary_file_sets_correct_chmod(self):
        manager = TempManager()
        path = manager.assign_temporary_file()

        try:
            st = os.stat(path)

            self.assertEqual('0755', str(oct(st.st_mode))[-4:])
        finally:
            manager.finally_clean_up()

    def test_chdir_is_considered(self):
        manager = TempManager('/tmp/')
        self.assertTrue(manager.create_tmp_file_path()[0].startswith('/tmp/'))

    def test_directory_is_created_when_it_does_not_exist(self):
        temp_chdir = '/tmp/' + str(uuid4())
        manager = TempManager(temp_chdir)

        try:
            manager.assign_temporary_file()

            self.assertTrue(os.path.isdir(temp_chdir))
        finally:
            manager.finally_clean_up()
            subprocess.call(['rmdir', temp_chdir])



#!/usr/bin/env python3

import unittest
from io import StringIO
from io import BytesIO
from rkd.api.inputoutput import StandardOutputReplication


class TestStandardOutputReplication(unittest.TestCase):
    def test_writing_to_multiple_streams_of_different_types(self):
        """Check that it is possible to write to multiple streams at once"""

        bytes_io = BytesIO()
        str_io = StringIO()

        out = StandardOutputReplication([bytes_io, str_io])
        out.write('12 June 1963 Medgar Wiley Evers, African American civil rights activist from Mississippi')
        out.write(' was shot in the back and killed by a member of the White Citizens\' Council.')

        for output in [bytes_io, str_io]:
            self.assertIn('Medgar Wiley Evers', str(output.getvalue()))
            self.assertIn('civil rights activist from Mississippi was shot in the back and killed',
                          str(output.getvalue()))

    def test_non_printable_is_converted_to_string_like_print_does(self):
        """Assert that non-str is casted to str, like regular print does. So the method is bulletproof"""

        str_io = StringIO()

        out = StandardOutputReplication([str_io])
        out.write(unittest)

        self.assertIn("<module 'unittest'", str_io.getvalue())




import pkg_resources
import os
import re
from subprocess import CalledProcessError
from typing import Dict
from typing import List
from argparse import ArgumentParser
from typing import Callable
from typing import Optional
from copy import deepcopy
from ..api.contract import TaskInterface
from ..api.contract import ExecutionContext
from ..api.contract import TaskDeclarationInterface
from ..api.contract import ArgparseArgument
from ..inputoutput import SystemIO
from ..inputoutput import clear_formatting
from ..aliasgroups import parse_alias_groups_from_env, AliasGroup


class InitTask(TaskInterface):
    """
    :init task is executing ALWAYS. That's a technical, core task.

    The purpose of this task is to handle global settings
    """

    def get_name(self) -> str:
        return ':init'

    def get_group_name(self) -> str:
        return ''

    def get_declared_envs(self) -> Dict[str, str]:
        return {
            'RKD_DEPTH': '0',
            'RKD_PATH': '',
            'RKD_ALIAS_GROUPS': '',
            'RKD_UI': 'true'
        }

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--no-ui', '-n', action='store_true',
                            help='Do not display RKD interface (similar to --silent, ' +
                                 'but does not inherit --silent into next tasks)')

    def execute(self, context: ExecutionContext) -> bool:
        """
        :init task is setting user-defined global defaults on runtime
        It allows user to call eg. rkd --log-level debug :task1 :task2
        to set global settings such as log level

        :param context:
        :return:
        """

        # increment RKD_DEPTH
        os.environ['RKD_DEPTH'] = str(int(os.getenv('RKD_DEPTH', '0')) + 1)

        self._ctx.io  # type: SystemIO
        self._ctx.io.silent = context.args['silent']

        # log level is optional to be set
        if context.args['log_level']:
            self._ctx.io.set_log_level(context.args['log_level'])

        if context.get_env('RKD_UI'):
            self._ctx.io.set_display_ui(context.get_env('RKD_UI').lower() == 'true')

        if int(os.getenv('RKD_DEPTH')) >= 2 or context.args['no_ui']:
            self._ctx.io.set_display_ui(False)

        return True

    def is_silent_in_observer(self) -> bool:
        return True


class TasksListingTask(TaskInterface):
    """Lists all enabled tasks

    Environment:
        - RKD_WHITELIST_GROUPS: Comma separated list of groups that should be only visible, others would be hidden
    """

    def get_name(self) -> str:
        return ':tasks'

    def get_group_name(self) -> str:
        return ''

    def configure_argparse(self, parser: ArgumentParser):
        pass

    def get_declared_envs(self) -> Dict[str, str]:
        return {
            'RKD_WHITELIST_GROUPS': '',
            'RKD_ALIAS_GROUPS': ''
        }

    def execute(self, context: ExecutionContext) -> bool:
        io = self._io
        groups = {}
        aliases = parse_alias_groups_from_env(context.get_env('RKD_ALIAS_GROUPS'))

        # fancy stuff
        whitelisted_groups = context.get_env('RKD_WHITELIST_GROUPS').replace(' ', '').split(',') \
            if context.get_env('RKD_WHITELIST_GROUPS') else []

        # collect into groups
        for name, declaration in self._ctx.find_all_tasks().items():
            group_name = declaration.get_group_name()

            # (optional) whitelists of displayed groups
            if whitelisted_groups:
                group_to_whitelist_check = (':' + group_name) if group_name else ''  # allow empty group ([global])

                if group_to_whitelist_check not in whitelisted_groups:
                    continue

            if group_name not in groups:
                groups[group_name] = {}

            groups[group_name][self.translate_alias(declaration.to_full_name(), aliases)] = declaration

        # iterate over groups and list tasks under groups
        for group_name, tasks in groups.items():
            if not group_name:
                group_name = 'global'

            io.print_group(group_name)

            for task_name, declaration in tasks.items():
                declaration: TaskDeclarationInterface

                try:
                    description = declaration.get_description()
                    text_description = "# " + description if description else ""
                except AttributeError:
                    text_description = ""

                io.outln(
                    self.ljust_task_name(declaration, task_name)
                    + text_description
                )

            io.print_opt_line()

        io.print_opt_line()
        io.opt_outln('Use --help to see task environment variables and switches, eg. rkd :sh --help, rkd --help')

        return True

    @staticmethod
    def translate_alias(full_name: str, aliases: List[AliasGroup]) -> str:
        if not aliases:
            return full_name

        for alias in aliases:
            match = alias.get_aliased_task_name(full_name)

            if match:
                return match

        return full_name

    @staticmethod
    def ljust_task_name(declaration: TaskDeclarationInterface, task_name: str) -> str:
        with_fancy_formatting = declaration.format_task_name(task_name)
        without_fancy_formatting = clear_formatting(with_fancy_formatting)
        ljust_only = without_fancy_formatting.ljust(50, ' ')[len(without_fancy_formatting):]

        return with_fancy_formatting + ljust_only


class CallableTask(TaskInterface):
    """ Executes a custom callback - allows to quickly define a short task """

    _callable: Callable[[ExecutionContext, TaskInterface], bool]
    _args_callable: Callable[[ArgumentParser], None]
    _argparse_options: Optional[List[ArgparseArgument]]
    _name: str
    _group: str
    _description: str
    _envs: dict
    _become: str

    def __init__(self, name: str, callback: Callable[[ExecutionContext, TaskInterface], bool],
                 args_callback: Callable[[ArgumentParser], None] = None,
                 description: str = '',
                 group: str = '',
                 become: str = '',
                 argparse_options: List[ArgparseArgument] = None):
        self._name = name
        self._callable = callback
        self._args_callable = args_callback
        self._description = description
        self._group = group
        self._envs = {}
        self._become = become
        self._argparse_options = argparse_options

    def get_name(self) -> str:
        return self._name

    def get_become_as(self) -> str:
        return self._become

    def get_description(self) -> str:
        return self._description

    def get_group_name(self) -> str:
        return self._group

    def configure_argparse(self, parser: ArgumentParser):
        if self._argparse_options:
            for opts in self._argparse_options:
                parser.add_argument(*opts.args, **opts.kwargs)

        if self._args_callable:
            self._args_callable(parser)

    def execute(self, context: ExecutionContext) -> bool:
        return self._callable(context, self)

    def push_env_variables(self, envs: dict):
        self._envs = deepcopy(envs)

    def get_declared_envs(self) -> Dict[str, str]:
        return self._envs


class VersionTask(TaskInterface):
    """ Shows version of RKD and of all loaded tasks """

    def get_name(self) -> str:
        return ':version'

    def get_group_name(self) -> str:
        return ''

    def configure_argparse(self, parser: ArgumentParser):
        pass

    def execute(self, context: ExecutionContext) -> bool:
        self._io.outln('RKD version %s' % pkg_resources.get_distribution("rkd").version)
        self._io.print_opt_line()

        table_body = []

        for name, declaration in self._ctx.find_all_tasks().items():
            if not isinstance(declaration, TaskDeclarationInterface):
                continue

            task = declaration.get_task_to_execute()
            class_name = str(task.__class__)
            module = task.__class__.__module__
            parts = module.split('.')

            for num in range(0, len(parts) + 1):
                try_module_name = ".".join(parts)

                try:
                    version = pkg_resources.get_distribution(try_module_name).version
                    table_body.append([name, version, module, class_name])

                    break
                except pkg_resources.DistributionNotFound:
                    parts = parts[:-1]
                except ValueError:
                    table_body.append([name, 'UNKNOWN (local module?)', module, class_name])

        self.io().outln(self.table(
            header=['Name', 'Version', 'Imported from', 'Representation'],
            body=table_body
        ))

        return True


class LineInFileTask(TaskInterface):
    """Adds or removes a line in file (works similar to the lineinfile from Ansible)
    """

    def get_name(self) -> str:
        return ':line-in-file'

    def get_group_name(self) -> str:
        return ':file'

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('file', help='File name')
        parser.add_argument('--output', help='Output filename (optional, when not specified, ' +
                                             'then source would be overwritten)', default='', required=False)
        parser.add_argument('--regexp', '-r', help='Regexp to find existing occurrence of line', required=True)
        parser.add_argument('--insert', '-i', help='Line to insert or replace', required=True)
        parser.add_argument('--new-after-line', '-ir', help='Regexp to match a line ' +
                                                            'after which a new entry should be added (often called ' +
                                                            'a header or marker line below which we want to add a text)')
        parser.add_argument('--only-first-occurrence', help='Stop at first occurrence', action='store_true')
        parser.add_argument('--fail-on-no-occurrence', '-f', help='Return 1 exit code, when no any occurrence found',
                            action='store_true')

    def execute(self, context: ExecutionContext) -> bool:
        file = context.get_arg('file')
        output_file = context.get_arg('--output') if context.get_arg('--output') else file
        line = context.get_arg('--insert').split("\n")
        regexp = context.get_arg('--regexp')
        after_line_regexp = context.get_arg('--new-after-line')
        fail_on_no_occurrence = context.get_arg('--fail-on-no-occurrence')
        only_first_occurrence = bool(context.get_arg('--only-first-occurrence'))

        if not os.path.isfile(output_file):
            self.io().info('Creating empty file "%s"' % output_file)
            self.sh('touch "%s"' % output_file)

        with open(file, 'r') as f:
            content = f.readlines()

        new_contents = ""
        found = 0

        for file_line in content:
            match = re.match(regexp, file_line)

            if match:
                self.io().debug('Found occurrence of "%s"' % regexp)

                if found and only_first_occurrence:
                    new_contents += file_line
                    continue

                file_line = ("\n".join(line)) + "\n"

                found += 1
                group_num = 0

                for value in list(match.groups()):
                    var_to_replace = '$match[' + str(group_num) + ']'
                    self.io().debug('Replacing "%s" with "%s"' % (var_to_replace, value))
                    file_line = file_line.replace(var_to_replace, value)
                    group_num += 1

            new_contents += file_line

        if not found:
            if fail_on_no_occurrence:
                self.io().error_msg('No matching line for selected regexp found')
                return False

        new_contents = self._insert_new_lines_if_necessary(found, new_contents, line, after_line_regexp,
                                                           only_first_occurrence, regexp)

        with open(output_file, 'w') as f:
            f.write(new_contents)

        self.io().success_msg('Replaced %i lines with "%s" in "%s"' % (found, "\n".join(line), output_file))

        return True

    def _insert_new_lines_if_necessary(self, found_at_least_one_occurrence: bool, contents: str, lines_to_insert: list,
                                       after_line_regexp: str, only_first_occurrence: bool, regexp: str) -> str:
        """Inserts new lines (if necessary)

        For each MARKER (line after which we want to insert new lines)
        check if there are lines already, if not then insert them.
        """

        if not after_line_regexp:
            self.io().debug('No header (marker) regexp defined')

            if found_at_least_one_occurrence:
                return contents

            # if not found_at_least_one_occurrence
            return contents + ("\n".join(lines_to_insert)) + "\n"

        as_lines = contents.split("\n")

        new_lines = []
        current_line_num = -1
        inserted_already_only_first_occurrence_allowed = False

        for line in as_lines:
            current_line_num += 1
            new_lines.append(line)

            if inserted_already_only_first_occurrence_allowed:
                continue

            if re.match(after_line_regexp, line):
                self.io().debug('Matched header line: "%s"' % line)

                # try to skip insertion, if the line already exists (do not duplicate lines)
                # WARNING: Matches only two lines after marker, that's a limitation
                next_lines = "\n".join(as_lines[current_line_num + 1:current_line_num + 2])

                if next_lines and re.match(regexp, next_lines):
                    continue

                self.io().debug('Inserting')
                new_lines += lines_to_insert

                if only_first_occurrence:
                    self.io().debug('Only first occurrence - stopping there with insertions')
                    inserted_already_only_first_occurrence_allowed = True

        return "\n".join(new_lines)


class CreateStructureTask(TaskInterface):
    """ Creates a RKD file structure in current directory

This task is designed to be extended, see methods marked as "interface methods".
    """

    def get_name(self) -> str:
        return ':create-structure'

    def get_group_name(self) -> str:
        return ':rkd'

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--commit', '-c',
                            help='Prepare RKD structure as a git commit (needs workspace to be clean)',
                            action='store_true')
        parser.add_argument('--no-venv', help='Do not create virtual env automatically', action='store_true')
        parser.add_argument('--pipenv', help='Generate files for Pipenv', action='store_true')

    def execute(self, ctx: ExecutionContext) -> bool:
        commit_to_git = ctx.get_arg('--commit')
        without_venv = ctx.get_arg('--no-venv')
        use_pipenv = bool(ctx.get_arg('--pipenv'))

        if commit_to_git and not self.check_git_is_clean():
            self.io().error_msg('Current working directory is dirty, you have working changes, ' +
                                'please commit them or stash first')
            return False

        template_structure_path = os.path.dirname(os.path.realpath(__file__)) + '/../misc/initial-structure'

        self.on_startup(ctx)

        # 1) Create structure from template
        if not os.path.isdir('.rkd'):
            self._io.info_msg('Creating a folder structure at %s' % os.getcwd())
            self.sh('cp -pr %s/.rkd ./' % template_structure_path, verbose=True)
            self.on_files_copy(ctx)
        else:
            self.io().info_msg('Not creating .rkd directory, already present')

        # 2) Populate git ignore
        self.sh('touch .gitignore')

        for file_path in self.get_patterns_to_add_to_gitignore(ctx):
            self.rkd([':file:line-in-file',
                      '.gitignore',
                      '--regexp="%s"' % file_path.replace('*', '\*'),
                      '--insert=%s' % file_path
                      ])

        # 3) Add RKD to requirements
        self.io().info('Adding RKD to requirements.txt')
        self._write_requirements(ctx)

        # 4) Create virtual env
        if not without_venv:
            self.io().info('Setting up virtual environment')
            self.on_creating_venv(ctx)
            self._setup_venv(use_pipenv, template_structure_path)

        if commit_to_git:
            self.on_git_add(ctx)
            self.git_add('.gitignore')
            self.git_add('setup-venv.sh')
            self.git_add('requirements.txt')
            self.git_add('.rkd')

            self.commit_to_git()

        self.print_success_msg(use_pipenv, ctx)

        return True

    def _write_requirements(self, ctx: ExecutionContext):
        self.sh('touch requirements.txt')
        self.rkd([':file:line-in-file',
                  'requirements.txt',
                  '--regexp="rkd(.*)"',
                  '--insert="rkd%s"' % self.get_rkd_version_selector()
                  ])
        self.on_requirements_txt_write(ctx)

    def _setup_venv(self, use_pipenv: bool, template_structure_path: str):
        if use_pipenv:
            self.sh('pipenv install rkd%s' % self.get_rkd_version_selector())
            return

        self.sh('cp %s/setup-venv.sh ./' % template_structure_path)
        self.sh('chmod +x setup-venv.sh')
        self.sh('./setup-venv.sh')

    @staticmethod
    def get_rkd_version_selector():
        rkd_version = pkg_resources.get_distribution("rkd").version
        return '==%s' % rkd_version

    def on_requirements_txt_write(self, ctx: ExecutionContext) -> None:
        """After requirements.txt file is written

        Interface method: to be overridden
        """

        pass

    def get_patterns_to_add_to_gitignore(self, ctx: ExecutionContext) -> list:
        """List of patterns to write to .gitignore

        Interface method: to be overridden
        """

        return ['.rkd/logs', '*.pyc', '*__pycache__*', '/.venv', '.venv-setup.log']

    def on_startup(self, ctx: ExecutionContext) -> None:
        """When the command is triggered, and the git is not dirty

        Interface method: to be overridden
        """

        pass

    def on_files_copy(self, ctx: ExecutionContext) -> None:
        """When files are copied

        Interface method: to be overridden
        """

        pass

    def on_creating_venv(self, ctx: ExecutionContext) -> None:
        """When creating virtual environment

        Interface method: to be overridden
        """

        pass

    def on_git_add(self, ctx: ExecutionContext) -> None:
        """Action on, when adding files via `git add`

        Interface method: to be overridden
        """

        pass

    def print_success_msg(self, use_pipenv: bool, ctx: ExecutionContext) -> None:
        """Emits a success message

        Interface method: to be overridden
        """

        if use_pipenv:
            self.io().success_msg("Structure created, use \"pipenv shell\" to enter project environment\n" +
                                  "Add libraries, task providers, tools to the environment using \"pipenv install\"")
            return

        self.io().success_msg("Structure created, use eval $(./setup-venv.sh) to enter Python\'s " +
                              "virtual environment with installed desired RKD version from requirements.txt\n" +
                              "Add libraries, task providers, tools to the requirements.txt " +
                              "for reproducible environments")

    def commit_to_git(self):
        if not os.path.isdir('.git'):
            return

        try:
            self.sh('git commit -m "Create RKD structure"')
        except CalledProcessError as e:
            if 'nothing to commit, working tree clean' in e.output:
                return

            raise e

    def git_add(self, path: str):
        if not os.path.isdir('.git'):
            return

        self.sh('git add "%s"' % path)

    def check_git_is_clean(self):
        return self.sh('git diff --stat || true', capture=True).strip() == ''



#!/usr/bin/env python3

import unittest
from rkd.standardlib import TasksListingTask
from rkd.test import get_test_declaration


class TestTasksListingTask(unittest.TestCase):
    def test_ljust_task_name(self):
        """Assert that the formatting is not breaking the description alignment
        The formatting instructions should not be considered by ljust. Only visible characters should be considered
        in padding with spaces"""

        # colored
        colored_declaration = get_test_declaration()
        colored_declaration.format_task_name = lambda text: "\x1B[93m" + text + "\x1B[0m"
        ljusted_colored = TasksListingTask.ljust_task_name(colored_declaration, ':general-strike')

        # not colored
        regular_declaration = get_test_declaration()
        regular_declaration.format_task_name = lambda text: text
        ljusted_regular = TasksListingTask.ljust_task_name(regular_declaration, ':general-strike')

        # assert: the coloring should not impact on the filling up size
        self.assertTrue(ljusted_colored.endswith((' ' * 35)), msg='Expected 35 spaces at the end (fill up)')
        self.assertTrue(ljusted_regular.endswith((' ' * 35)), msg='Expected 35 spaces at the end (fill up)')



#!/usr/bin/env python3

import unittest
from rkd.test import get_test_declaration


class TestTaskDeclaration(unittest.TestCase):
    def test_get_full_description_returns_docstring_when_description_method_does_not_return_anything(self):
        """TaskDeclaration.get_full_description() returns docstring of task to execute, when the task does not implement
        get_description() by itself
        """

        declaration = get_test_declaration()
        declaration.get_task_to_execute().get_description = lambda: ''

        declaration.get_task_to_execute().__doc__ = '''
            10 Jun 1973 a strike of gravediggers at 3 cemeteries in the New York metropolitan area was expanded 
            to include 44 others. Strikers defied a court order, the jailing of their union leader and after 
            27 days they won a pension scheme and pay increase
        '''

        self.assertIn('27 days they won a pension scheme and pay increase', declaration.get_full_description())

    def test_get_full_description_returns_result_of_tasks_get_description_if_defined(self):
        """TaskDeclaration.get_full_description() returns get_task_to_execute().get_description() if response is not
        empty
        """

        declaration = get_test_declaration()
        declaration.get_task_to_execute().get_description = lambda: '10 June 1942 Polish prisoners on work detail ' + \
                                                                    'at Auschwitz Birkenau managed to escape while ' + \
                                                                    'digging a drainage ditch.'

        self.assertIn('Polish prisoners', declaration.get_full_description())
        self.assertIn('Auschwitz Birkenau', declaration.get_full_description())
        self.assertIn('managed to escape', declaration.get_full_description())

    def test_get_full_description_allows_empty_doc(self):
        """Test that description can be not defined at all"""

        declaration = get_test_declaration()
        declaration.get_task_to_execute().get_description = lambda: None
        declaration.get_task_to_execute().__doc__ = None

        self.assertEqual('', declaration.get_full_description())



#!/usr/bin/env python3

import unittest
from rkd.contract import env_to_switch
from rkd.contract import ArgumentEnv
from rkd.exception import EnvironmentVariableNameNotAllowed


class ArgumentEnvTest(unittest.TestCase):
    def test_env_to_switch(self):
        self.assertEqual('--michael-brown', env_to_switch('MICHAEL_BROWN'))

    def test_reserved_vars_are_not_allowed(self):
        reserved_vars = ['PATH', 'PWD', 'LANG', 'DISPLAY', 'SHELL', 'SHLVL', 'HOME', 'EDITOR']

        for var_name in reserved_vars:
            with self.subTest('Environment variable "%s" should be disallowed' % var_name):
                self.assertRaises(EnvironmentVariableNameNotAllowed,
                                  lambda: ArgumentEnv(name=var_name))

    def test_not_reserved_vars_are_allowed(self):
        not_reserved_examples = ['COMMAND', 'RKD_CMD']

        for var_name in not_reserved_examples:
            with self.subTest('Environment variable "%s" should be allowed to declare' % var_name):
                self.assertIsInstance(ArgumentEnv(name=var_name), ArgumentEnv)



import yaml
import os
import importlib
from types import FunctionType
from typing import List, Tuple, Union, Callable
from dotenv import dotenv_values
from copy import deepcopy
from collections import OrderedDict
from .exception import YamlParsingException
from .exception import EnvironmentVariablesFileNotFound
from .api.syntax import TaskDeclaration
from .api.syntax import TaskAliasDeclaration
from .api.contract import ExecutionContext
from .api.contract import TaskInterface
from .api.contract import ArgparseArgument
from .standardlib import CallableTask
from .api.inputoutput import IO
from .yaml_parser import YamlFileLoader
from .execution.declarative import DeclarativeExecutor


class YamlSyntaxInterpreter:
    """
    Translates YAML syntax into Python syntax of makefile (makefile.yaml -> makefile.py)
    """

    io: IO
    loader: YamlFileLoader

    def __init__(self, io: IO, loader: YamlFileLoader):
        self.io = io
        self.loader = loader

    def parse(self, content: str, rkd_path: str, file_path: str) \
            -> Tuple[List[TaskDeclaration], List[TaskAliasDeclaration]]:

        """ Parses whole YAML into entities same as in makefile.py - IMPORTS, TASKS """

        pre_parsed = yaml.load(content, yaml.Loader)

        if 'version' not in pre_parsed:
            raise YamlParsingException('"version" is not specified in YAML file')

        parsed = self.loader.load(content, pre_parsed['version'])

        imports = []
        global_envs = self.parse_env(parsed, file_path)

        if "imports" in parsed:
            imports = self.parse_imports(parsed['imports'])

        tasks = self.parse_tasks(
            parsed['tasks'] if 'tasks' in parsed else {},
            rkd_path,
            file_path,
            global_envs
        )

        return imports + tasks, []

    def parse_tasks(self, tasks: dict, rkd_path: str, makefile_path, global_envs: OrderedDict) -> List[TaskDeclaration]:
        """ Parse tasks section of YAML and creates rkd.standardlib.CallableTask type tasks

        Arguments:
            tasks: List of tasks
            rkd_path: Path to the .rkd directory
            makefile_path: Path to the makefile.yaml/yml that is being parsed
            global_envs: Environment defined in global scope of YAML document (WITHOUT os.environ)
        """

        parsed_tasks: List[Union[TaskDeclaration, None]] = []

        for name, yaml_declaration in tasks.items():
            parsed_tasks.append(self._parse_task(name, yaml_declaration, rkd_path, global_envs, makefile_path))

        return parsed_tasks

    def parse_env(self, parent: dict, makefile_path: str) -> OrderedDict:
        """Parse environment variables from parent node

        Priority (first - higher): 1) environment 2) env_files

        Examples:
            environment:
                EVENT_NAME: "In memory of Maxwell Itoya, an Nigerian immigrant killed by police at flea market. He was innocent, and left wife with 3 kids."

            env_files:
                - .rkd/env/important-history-events.envs

        Returns:
            KV dictionary
        """
        envs = OrderedDict()

        if "env_files" in parent:
            for path in parent['env_files']:
                envs.update(self._load_env_from_file(path, makefile_path))

        if "environment" in parent:
            envs.update(parent['environment'])

        return envs

    @staticmethod
    def _load_env_from_file(path: str, makefile_path: str) -> dict:
        """Load .env file

        Loads .env file in Bash-like syntax from selected path (path to file, not directory).
        Looks also in .rkd/{{path}} and in {{makefile_path}}/{{path}}

        Arguments:
            path: Path to file that contains env variables
            makefile_path: Path to the Makefile.yaml that is being processed

        Returns:
            KV dictionary
        """

        search_paths = [
            path,
            '.rkd/' + path,
            os.path.dirname(makefile_path) + '/' + path
        ]

        for search_path in search_paths:
            if not os.path.isfile(search_path):
                continue

            return dotenv_values(dotenv_path=search_path)

        raise EnvironmentVariablesFileNotFound(path, search_paths)

    def _parse_task(self, name: str, yaml_declaration: dict, rkd_path: str,
                    global_env: OrderedDict, makefile_path: str) -> TaskDeclaration:

        description = yaml_declaration['description'] if 'description' in yaml_declaration else ''
        arguments = yaml_declaration['arguments'] if 'arguments' in yaml_declaration else {}
        become = yaml_declaration['become'] if 'become' in yaml_declaration else ''

        # important: order of environment variables loading
        envs = deepcopy(global_env)
        envs.update(self.parse_env(yaml_declaration, makefile_path))
        envs.update(deepcopy(os.environ))

        try:
            steps = yaml_declaration['steps']

            # to make the syntax easier allow a single step
            if isinstance(steps, str):
                steps = [steps]

        except KeyError:
            raise YamlParsingException('"steps" are required to be defined in task "%s"' % name)

        task_name, group_name = TaskDeclaration.parse_name(name)

        return TaskDeclaration(
            CallableTask(
                name=task_name,
                description=description,
                group=group_name,
                argparse_options=self.parse_argparse_arguments(arguments),
                callback=self.create_execution_callback_from_steps(steps, name, rkd_path, envs),
                become=become
            )
        )

    @staticmethod
    def create_execution_callback_from_steps(steps: List[str], task_name: str, rkd_path: str, envs: dict) \
            -> Callable[[ExecutionContext, TaskInterface], bool]:

        """Creates implementation of TaskInterface.execute() - a callback that will execute all steps (callbacks)"""

        declarative = DeclarativeExecutor()

        #
        # Collect all steps and create micro-callbacks per step, that will be executed one-by-one in execute()
        #
        for step in steps:
            language = 'bash'
            as_lines = step.strip().split("\n")
            first_line = as_lines[0]
            code = step

            # code that begin with a hashbang will have hashbang cut off
            # #!bash
            # #!python
            if first_line[0:2] == '#!':
                language = first_line[2:]
                code = "\n".join(as_lines[1:])

            declarative.add_step(language, code, task_name, rkd_path, envs)

            if language not in ['python', 'bash']:
                raise YamlParsingException('Unsupported step language "%s"' % language)

        return declarative.execute_steps_one_by_one

    @staticmethod
    def parse_argparse_arguments(arguments: dict) -> List[ArgparseArgument]:
        """ Creates implementation of TaskInterface.configure_argparse() """

        converted = []

        for name, params in arguments.items():
            converted.append(ArgparseArgument([name], params))

        return converted

    @staticmethod
    def parse_imports(classes: List[str]) -> List[TaskDeclaration]:
        """Parses imports strings into Python classes

        Args:
            classes: List of classes to import

        Returns:
            A list of basic task declarations with imported tasks inside
            [
                TaskDeclaration(ProtestWorkplaceTask()),
                TaskDeclaration(StrikeWorkplaceTask()),
                TaskDeclaration(OccupyWorkplaceTask()),
                TaskDeclaration(RunProductionByWorkersOnTheirOwnTask())
            ]

        Raises:
            YamlParsingException: When a class or module does not exist
        """

        parsed: List[TaskDeclaration] = []

        for import_str in classes:
            parts = import_str.split('.')
            class_name = parts[-1]
            import_path = '.'.join(parts[:-1])

            # importing just a full module name eg. "rkd_python"
            if len(parts) == 1:
                import_path = import_str
                class_name = 'imports'
            # Test if it's not a class name
            # In this case we treat is as a module and import an importing method imports()
            elif class_name.lower() == class_name:
                import_path += '.' + class_name
                class_name = 'imports'

            try:
                module = importlib.import_module(import_path)
            except ImportError as e:
                raise YamlParsingException('Import "%s" is invalid - cannot import class "%s" - error: %s' % (
                    import_str, class_name, str(e)
                ))

            if class_name not in dir(module):
                raise YamlParsingException('Import "%s" is invalid. Class "%s" not found in module "%s"' % (
                    import_str, class_name, import_path
                ))

            if isinstance(module.__getattribute__(class_name), FunctionType):
                parsed += module.__getattribute__(class_name)()
            else:
                parsed.append(TaskDeclaration(module.__getattribute__(class_name)()))

        return parsed



#!/usr/bin/env python3

import sys
import unittest
from rkd.api.inputoutput import BufferedSystemIO
from rkd.api.inputoutput import output_formatted_exception
from rkd.api.inputoutput import indent_new_lines


class TestIOErrorHandling(unittest.TestCase):
    def test_is_information_written_through_stderr_methods(self):
        """To check if information is written to the proper methods we will mock stdout methods"""

        io = BufferedSystemIO()
        io._stdout = lambda *args, **kwargs: None

        try:
            raise IndexError('Invalid index 5')
        except Exception as exc:
            output_formatted_exception(exc, ':my-test-task', io)

        self.assertIn('IndexError', io.get_value())
        self.assertIn('Invalid index 5', io.get_value())
        self.assertIn('Retry with "-rl debug" switch before failed task to see stacktrace', io.get_value())

    def test_are_chained_exceptions_printed(self):
        """Check if there are multiple exceptions, then those would be printed as a cause"""

        io = BufferedSystemIO()

        try:
            try:
                raise IndexError('Invalid index 5')
            except IndexError as index_exc:
                raise Exception('There was an error with index') from index_exc

        except Exception as exc:
            output_formatted_exception(exc, ':my-test-task', io)

        self.assertIn('(Caused by) IndexError:', io.get_value())
        self.assertIn('Exception:', io.get_value())
        self.assertIn('There was an error with index', io.get_value())

    def test_fatal_error_would_be_thrown_in_case_of_a_formatting_failure(self):
        """Tests very important functionality - in case of a failure in the error formatting there must be a fallback
        that will work for sure - a simple print() of stacktrace and proper exit code
        """

        def mock_fatal(*args, **kwargs):
            raise Exception('Fatal!')

        io = BufferedSystemIO()
        io.print_line = mock_fatal

        exit_code = 0

        try:
            raise IndexError('Invalid index 5')
        except Exception as exc:
            stdout_bckp = sys.stdout  # needed to silence the output in tests (we do not want this stack trace in tests)

            try:
                with open('/dev/null', 'w') as sys.stdout:
                    output_formatted_exception(exc, ':my-test-task', io)

            except SystemExit as sys_exit:
                exit_code = sys_exit.code
            finally:
                sys.stdout = stdout_bckp  # restore stdout

        self.assertEqual(1, exit_code)

    def test_indent_new_lines(self):
        self.assertIn("\n      ", indent_new_lines("Line1\nLine2", 6))




import os
import sys
from typing import Union
from subprocess import check_output, Popen, DEVNULL, CalledProcessError
from abc import ABC as AbstractClass, abstractmethod
from .api.inputoutput import IO
from .process import check_call


class TaskUtilities(AbstractClass):
    """
    Internal helpers for TaskInterface implementations
    """

    @abstractmethod
    def io(self) -> IO:
        pass

    def silent_sh(self, cmd: str, verbose: bool = False, strict: bool = True,
                  env: dict = None) -> bool:
        """
        sh() shortcut that catches errors and displays using IO().error_msg()
        """

        # kwargs is not used on purpose. For static analysis.

        try:
            self.sh(cmd=cmd, capture=False, verbose=verbose, strict=strict, env=env)
            return True

        except CalledProcessError as e:
            self.io().error_msg(str(e))
            return False

    @staticmethod
    def get_rkd_binary():
        """Gets the command how RKD was launched"""

        if os.getenv('RKD_BIN'):
            return os.getenv('RKD_BIN')

        binary = sys.argv[0]
        sys_executable_basename = os.path.basename(sys.executable)

        if binary == 'python3 -m unittest':
            return 'python3 -m rkd'

        # as a Python module: "python -m rkd" for example
        if binary[:-3] == '.py':
            return '%s -m %s' % (sys.executable, os.path.basename(os.path.dirname(binary)))

        if sys_executable_basename.startswith('python'):
            return binary

        # using a script eg. "rkd"
        return sys.executable

    def sh(self, cmd: str, capture: bool = False, verbose: bool = False, strict: bool = True,
           env: dict = None) -> Union[str, None]:
        """ Executes a shell script in bash. Throws exception on error.
            To capture output set capture=True
        """

        self.io().debug('sh(%s)' % cmd)

        cmd = 'export PYTHONUNBUFFERED=1; ' + cmd

        # set strict mode, it can be disabled manually
        if strict:
            cmd = 'set -euo pipefail; ' + cmd

        if verbose:
            cmd = 'set -x; ' + cmd

        # append environment variables in order
        if env:
            env_str = ""

            for name, value in env.items():
                value = '' if value is None else str(value).replace('"', '\\"')
                env_str = env_str + (" export %s=\"%s\";\n" % (name, value))

            cmd = env_str + cmd

        bash_script = "#!/bin/bash -eopipefail \n" + cmd
        bash_script = bash_script.replace('%RKD%', self.get_rkd_binary())

        read, write = os.pipe()
        os.write(write, bash_script.encode('utf-8'))
        os.close(write)

        if not capture:
            check_call('bash', stdin=read, script=bash_script)
            return

        return check_output('bash', shell=True, stdin=read).decode('utf-8')

    def py(self, code: str, become: str = None, capture: bool = False, script_path: str = None) -> Union[str, None]:
        """Executes a Python code in a separate process"""

        read, write = os.pipe()
        os.write(write, code.encode('utf-8'))
        os.close(write)

        cmd = 'python'

        if script_path:
            cmd += ' ' + script_path + ' '

        if become:
            cmd = "sudo -E -u %s %s" % (become, cmd)

        os.putenv('RKD_BIN', self.get_rkd_binary())

        if not capture:
            check_call(cmd, stdin=read, script=code)
            return

        return check_output(cmd, shell=True, stdin=read).decode('utf-8')

    def exec(self, cmd: str, capture: bool = False, background: bool = False) -> Union[str, None]:
        """ Starts a process in shell. Throws exception on error.
            To capture output set capture=True
        """

        if background:
            if capture:
                raise Exception('Cannot capture output from a background process')

            Popen(cmd, shell=True, stdout=DEVNULL, stderr=DEVNULL)
            return

        if not capture:
            check_call(cmd)
            return

        return check_output(cmd, shell=True).decode('utf-8')

    def rkd(self, args: list, verbose: bool = False, capture: bool = False) -> str:
        """ Spawns an RKD subprocess
        """

        bash_opts = 'set -x; ' if verbose else ''
        args_str = ' '.join(args)

        return self.sh(bash_opts + ' %%RKD%% --no-ui %s' % args_str, capture=capture)



#!/usr/bin/env python3

import unittest
import os
from rkd.standardlib.jinja import FileRendererTask
from rkd.test import mock_task, mock_execution_context
from rkd.api.inputoutput import BufferedSystemIO

SAMPLES_PATH = os.path.dirname(os.path.realpath(__file__)) + '/internal-samples'


class TestFileRendererTask(unittest.TestCase):
    """Tests for a task that should render single JINJA2 file from SOURCE PATH to TARGET PATH
    """

    @staticmethod
    def _execute_mocked_task(params: dict, envs: dict = {}) -> BufferedSystemIO:
        io = BufferedSystemIO()

        task: FileRendererTask = FileRendererTask()
        mock_task(task, io=io)
        task.execute(mock_execution_context(task, params, envs))

        return io

    def test_naming(self):
        self.assertEqual(':j2:render', FileRendererTask().get_full_name())

    def test_renders_file_considering_environment_variables(self):
        """Test that file is properly rendered
        """

        msg = '22 May 1918 Spanish civil war fighter Dolores Jiménez Alvarez was born. Fought in ' + \
              'the Durruti column until arrested by the Communists. Then escaped and joined French ' + \
              'resistance and underground resistance to Franco'

        io = self._execute_mocked_task(
            # shell arguments
            {
                'source': SAMPLES_PATH + '/jinja2/example.j2',
                'output': '-'
            },
            # env variables
            {
                'MESSAGE': msg
            }
        )

        self.assertEqual('Listen to this important message: `%s`\n' % msg, io.get_value())

    def test_non_existing_source_file_raises_error(self):
        """Test that error message will be printed in case, when source file does not exist
        """

        io = self._execute_mocked_task(
            {
                'source': 'non-existing',
                'output': '-'
            }
        )

        self.assertIn('Source file does not exist at path "non-existing"', io.get_value())

    def test_undefined_variable_raises_error(self):
        """Test that JINJA2 will raise an exception, and it will be formatted properly
        """
        io = self._execute_mocked_task(
            # shell arguments
            {
                'source': SAMPLES_PATH + '/jinja2/example.j2',
                'output': '-'
            },
            # env variables
            {
                # "MESSAGE" variable should be defined, but is not - to test if error will be raised
            }
        )

        self.assertIn("Undefined variable - 'MESSAGE' is undefined", io.get_value())
        self.assertNotIn('Traceback (most recent call last):', io.get_value())

    def test_renders_with_extends_section(self):
        """Test that JINJA2 finds files on disk in current directory when using {% extends "..." %}"""

        cwd_copy = os.getcwd()
        os.chdir('../test/internal-samples/jinja2')

        try:
            io = self._execute_mocked_task(
                # shell arguments
                {
                    'source': SAMPLES_PATH + '/jinja2/example-with-extends.j2',
                    'output': '-'
                },
                # env variables
                {
                }
            )
        finally:
            os.chdir(cwd_copy)

        self.assertIn('Teresa Claramunt was born.', io.get_value())
        self.assertIn('played a leading role in the 1911 Aragon general strike.', io.get_value())
        self.assertIn('When she died 50k workers went to her funeral', io.get_value())
